from ..widget import Widget
from .layout import RelativeLayout


class SliderPreview(RelativeLayout):

    def __init__(
        self,
        pos=[0., 0.],
        size=[100., 10.],
        relative={},
        halign='left',
        valign='bottom',
        orientation='horizontal',
        bar_width=.3,
        bar_color=(1, 1, 1, 1),
        bar_hover_color=(1, 1, 1, 1),
        knob_size=1,
        knob_color=(.8, .8, .8, 1),
        knob_hover_color=(.8, .8, .8, 1),
        steps=-1,
        allow_bar_click=True,
        angle=0
    ):
        self.bar = None
        self.knob = None
        self._bar_width = bar_width
        self.bar = RelativeLayout(
            pos=(.5, .5),
            size=(0, 0),
            valign='center',
            halign='center',
            relative={'size': True, 'pos': True},
            bg_color=bar_color
        )
        self._orientation = orientation
        super().__init__(pos, size, (0, 0, 0, 0), relative, halign=halign, valign=valign, angle=angle)
        self._value = 0
        self.add_widget(self.bar)
        self.knob = RelativeLayout(pos=(.5, .5), relative={'pos': True}, size=(0, 0), bg_color=knob_color, valign='center', halign='center')
        self.bar.add_widget(self.knob)
        self.bar_width = bar_width
        self.knob_size = knob_size

    @property
    def bar_width(self):
        return self.bar.height if self.is_horizontal else self.bar.width

    @bar_width.setter
    def bar_width(self, val):
        self.bar.size = (1, val) if self.is_horizontal else (val, 1)
        self._bar_width = val

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        self._value = val
        self.knob.pos = (val, .5) if self.is_horizontal else (.5, val)

    @property
    def bar_color(self):
        return self.bar.bg_color

    @bar_color.setter
    def bar_color(self, val):
        self.bar.bg_color = val
        self.bar.click_color = val

    @property
    def bar_hover_color(self):
        return self.bar.hover_color

    @bar_hover_color.setter
    def bar_hover_color(self, val):
        self.bar.hover_color = val

    @property
    def knob_color(self):
        return self.knob.bg_color

    @knob_color.setter
    def knob_color(self, val):
        self.knob.bg_color = val
        self.knob.click_color = val

    @property
    def knob_hover_color(self):
        return self.knob.hover_color

    @knob_hover_color.setter
    def knob_hover_color(self, val):
        self.knob.hover_color = val

    @property
    def knob_size(self):
        return self.knob.size[0]

    @knob_size.setter
    def knob_size(self, val):
        fac = self._draw_size[1] if self.is_horizontal else self._draw_size[0]
        self.knob.size = (val * fac, val * fac)

    @property
    def border_width(self):
        return 0

    @border_width.setter
    def border_width(self, val):
        self.bar.border_width = 0

    @property
    def border_color(self):
        return self.bar.border_color

    @border_color.setter
    def border_color(self, val):
        self.bar.border_color = (0, 0, 0, 0)

    @property
    def orientation(self):
        return self._orientation
    
    @orientation.setter
    def orientation(self, val):
        self.bar_width = self._bar_width
        self._orientation = val

    @property
    def is_horizontal(self):
        return self.orientation != 'vertical'

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, val):
        if self.parent is not val and self.parent:
            self.parent.remove_widget(self)
        if self.use_clipping is None:
            self.use_clipping = val.use_clipping
        self._parent = val
        self.pos = self.pos
        self.size = self.size
        self._build_shader()

    def _build_shader(self):
        super()._build_shader()
        if self.bar and self.knob:
            self.bar._build_shader()
            self.knob._build_shader()


class FrameSliderPreview(SliderPreview):

    def __init__(
        self,
        pos=[0., 0.],
        size=[100., 10.],
        relative={},
        halign='left',
        valign='bottom',
        orientation='horizontal',
        border_width=1,
        border_color=(.8, .8, .8, 1),
        bg_color=(0, 0, 0, 0),
        bar_color=(1, 1, 1, 1),
        bar_hover_color=(.1, .1, .1, .3),
        steps=-1,
        allow_bar_click=True,
        angle=0
    ):
        self.bar = None
        self.knob = None
        Widget.__init__(self, pos, size, (0, 0, 0, 0), relative, halign=halign, valign=valign, angle=angle)
        self._value = 0
        # self._horiz = orientation != 'vertical'
        self._bar_width = 1
        self._orientation = orientation
        self.bar = RelativeLayout(
            pos=(.5, .5),
            size=(1, 1),
            valign='center',
            halign='center',
            relative={'size': True, 'pos': True},
            border_color=border_color,
            bg_color=bg_color,
        )
        self.border_color = border_color
        self.border_width = border_width
        self.knob = RelativeLayout(pos=(.5, .5), relative={'pos': True}, size=(1, 1), bg_color=bar_color, valign='center', halign='center')
        self.knob_size = 1
        self.add_widget(self.bar)
        self.add_widget(self.knob)
        self.start()

    @property
    def bar_width(self):
        return 1

    @bar_width.setter
    def bar_width(self, val):
        self.bar.size = (1, 1)

    @property
    def border_width(self):
        return self.bar.border_width

    @border_width.setter
    def border_width(self, val):
        self.bar.border_width = val

    @property
    def border_color(self):
        return self.bar.border_color

    @border_color.setter
    def border_color(self, val):
        self.bar.border_color = val

    @property
    def knob_size(self):
        return self.knob.size[0]

    @knob_size.setter
    def knob_size(self, val):
        fac = self._draw_size[1] if self.is_horizontal else self._draw_size[0]
        self.knob.size = (fac, fac) if self.is_horizontal else (fac, fac)


class ProgressSliderPreview(FrameSliderPreview):

    def __init__(self, pos=[0, 0], size=[100, 10], relative={}, halign='left', valign='bottom', orientation='horizontal', border_width=1, border_color=(0.8, 0.8, 0.8, 1), bg_color=(0, 0, 0, 0), bar_color=(1, 1, 1, 1), bar_hover_color=(0.1, 0.1, 0.1, 0.3), steps=-1, allow_bar_click=True, angle=0):
        self.bar = None
        self.knob = None
        self._orientation = orientation
        super().__init__(pos, size, relative, halign, valign, orientation, border_width, border_color, bg_color, bar_color, bar_hover_color, steps, allow_bar_click, angle)
        self.knob.relative = {'size': True}
        self.knob.halign = 'left'
        self.knob.valign = 'bottom'
        self.start()

    @property
    def knob_size(self):
        return (1, 1)
        # return self.knob.size[0]

    @knob_size.setter
    def knob_size(self, val):
        pass
        # self.knob.size = (.5, 1) if self.is_horizontal else (1, .5)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        self._value = val
        self.knob.size = (val, 1) if self.is_horizontal else (1, val)



from uplogic.nodes import ULParameterNode


class ULColorRGB(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.color = None
        self.output_vector = None
        self.OUTV = self.add_output(self.get_out_v)

    def get_out_v(self):
        c = self.get_input(self.color)
        c = c.copy()
        c.resize_3d()
        return c



from uplogic.nodes import ULConditionNode
from bge import logic


class ULMouseReleased(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False
        self.mouse_button_code = None
        self.network = None
        self.OUT = self.add_output(self.get_changed)

    def get_changed(self):
        mouse_button = self.get_input(self.mouse_button_code)
        mstat = logic.mouse.inputs[mouse_button]
        if self.pulse:
            return (
                    mstat.released or
                    mstat.inactive
                )
        else:
            return (mstat.released)

    def setup(self, network):
        self.network = network



from uplogic.nodes import ULActionNode


class ULRunActuator(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_obj = None
        self.cont_name = None
        self.act_name = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        game_obj = self.get_input(self.game_obj)
        cont_name = self.get_input(self.cont_name)
        act_name = self.get_input(self.act_name)
        controller = game_obj.controllers[cont_name]
        if act_name not in controller.actuators:
            return
        actuator = controller.actuators[act_name]
        if not self.get_condition():
            return
        controller.activate(actuator)
        self._done = True



from uplogic.nodes import ULActionNode
from uplogic.ui import Slider, FrameSlider, ProgressSlider
from mathutils import Vector
from math import degrees


class ULCreateUISlider(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.parent = None
        self.rel_pos = None
        self.pos = None
        self.rel_size = None
        self.size = None
        self.relative = None
        self.angle = None
        self.bar_width = None
        self.border_width = None
        self.border_color = None
        self.bar_color = None
        self.bar_hover_color = None
        self.knob_size = None
        self.knob_color = None
        self.knob_hover_color = None
        self.steps = None
        self.initial_value = 0
        self.allow_bar_click = None
        self._widget = None
        self.halign_type = 'left'
        self.valign_type = 'bottom'
        self.orientation_type = 'horizontal'
        self.slider_type = '0'
        self.OUT = self.add_output(self._get_done)
        self.WIDGET = self.add_output(self._get_widget)
        self.VALUE = self.add_output(self._get_value)
        self.KNOB_POSITION = self.add_output(self._get_knob_pos)

    def _get_done(self):
        return self._done

    def _get_widget(self):
        return self._widget

    def _get_value(self):
        return self._widget.value if self._widget else 0

    def _get_knob_pos(self):
        return Vector(self._widget.knob.pos_abs) if self._widget else Vector((0, 0))

    def evaluate(self):
        if not self.get_condition():
            return
        ipt = self.get_input

        parent = ipt(self.parent)

        if self.slider_type == '2':
            self._widget = ProgressSlider(
                pos=ipt(self.pos),
                size=ipt(self.size),
                relative={'pos': ipt(self.rel_pos), 'size': ipt(self.rel_size)},
                halign=ipt(self.halign_type),
                valign=ipt(self.valign_type),
                orientation=ipt(self.orientation_type),
                border_width=ipt(self.border_width),
                border_color=ipt(self.border_color),
                bg_color=ipt(self.bar_color),
                bar_color=ipt(self.knob_color),
                bar_hover_color=ipt(self.bar_hover_color),
                steps=ipt(self.steps),
                allow_bar_click=ipt(self.allow_bar_click),
                angle=degrees(ipt(self.angle))
            )
        elif self.slider_type == '1':
            self._widget = FrameSlider(
                pos=ipt(self.pos),
                size=ipt(self.size),
                relative={'pos': ipt(self.rel_pos), 'size': ipt(self.rel_size)},
                halign=ipt(self.halign_type),
                valign=ipt(self.valign_type),
                orientation=ipt(self.orientation_type),
                border_width=ipt(self.border_width),
                border_color=ipt(self.border_color),
                bg_color=ipt(self.bar_color),
                bar_color=ipt(self.knob_color),
                bar_hover_color=ipt(self.bar_hover_color),
                steps=ipt(self.steps),
                allow_bar_click=ipt(self.allow_bar_click),
                angle=degrees(ipt(self.angle))
            )
        else:
            self._widget = Slider(
                pos=ipt(self.pos),
                size=ipt(self.size),
                relative={'pos': ipt(self.rel_pos), 'size': ipt(self.rel_size)},
                halign=ipt(self.halign_type),
                valign=ipt(self.valign_type),
                orientation=ipt(self.orientation_type),
                bar_width=ipt(self.bar_width),
                # border_color=ipt(self.border_color),
                bar_color=ipt(self.bar_color),
                bar_hover_color=ipt(self.bar_hover_color),
                knob_size=ipt(self.knob_size),
                knob_color=ipt(self.knob_color),
                knob_hover_color=ipt(self.knob_hover_color),
                steps=ipt(self.steps),
                allow_bar_click=ipt(self.allow_bar_click),
                angle=degrees(ipt(self.angle))
            )
        self._widget.value = self.get_input(self.initial_value)
        if parent:
            parent.add_widget(self._widget)
        self._done = True
        



from uplogic.nodes import ULActionNode
from uplogic.shaders import remove_filter


class ULRemoveFilter(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.pass_idx = None
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        pass_idx = self.get_input(self.pass_idx)
        remove_filter(pass_idx)
        self._done = True



from uplogic.nodes import ULActionNode
from uplogic.audio.fmod import FMod


class FModStartEventNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = False
        self.event = ''
        self.source = ''
        self.channel = 'default'
        self.mask = 65535
        self._evt = None
        self.OUT = self.add_output(self.get_done)
        self.EVT = self.add_output(self.get_evt)

    def get_done(self):
        return self._done

    def get_evt(self):
        return self._evt

    def evaluate(self):
        if not self.get_condition():
            return
        self._evt = FMod.event(
            f'event:/{self.get_input(self.event)}',
            self.get_input(self.source),
            self.get_input(self.channel)
        )
        self._evt.occlusion_mask = self.mask
        self._done = True



from uplogic.nodes import ULParameterNode


class GetCollisionGroupNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.mode = 0
        self.INT = self.add_output(self.get_int)

    def get_int(self):
        game_object = self.get_input(self.game_object)
        return game_object.collisionMask if self.mode else game_object.collisionGroup



from uplogic.nodes import ULConditionNode
from uplogic import console


class ULControllerStatus(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.obj_name = None
        self.cont_name = None
        self.OUT = self.add_output(self.get_controller)
        self.SENSORS = self.add_output(self.get_sensors)

    def get_controller(self):
        game_obj = self.game_obj
        cont = game_obj.controllers[self.controller]
        state = (
            game_obj
            .blenderObject
            .game
            .controllers[self.controller]
            .type
        )
        if not cont.sensors:
            return False
        elif state == 'LOGIC_AND':
            return False not in [sens.positive for sens in cont.sensors]
        elif state == 'LOGIC_OR':
            return True in [sens.positive for sens in cont.sensors]
        elif state == 'LOGIC_NAND':
            return False in [sens.positive for sens in cont.sensors]
        elif state == 'LOGIC_NOR':
            return True not in [sens.positive for sens in cont.sensors]
        elif state == 'LOGIC_XOR':
            return [
                    sens.positive
                    for sens in
                    cont.sensors
                ].count(True) % 2 != 0
        elif state == 'LOGIC_XNOR':
            check = cont.sensors[0].positive
            return False not in [
                    sens.positive == check
                    for sens in
                    cont.sensors
                ]
        else:
            console.debug('Expression/Python not supported for controller.')
            return False

    def get_sensors(self):
        sensors = {}
        cont = self.controller
        for sens in self.game_obj.controllers[cont].sensors:
            sensors[sens] = sens.positive
        return sensors

    def evaluate(self):
        self.game_obj = self.get_input(self.obj_name)
        self.controller = self.get_input(self.cont_name)



from bge import logic
from bge import constraints
from bpy.types import Collection as BColl
import bpy, bge
from mathutils import Vector, Euler
from uplogic import events


def assign(game_object: bge.types.KX_GameObject, collection: BColl, exclusive=True):
    """Link an object into a specified collection.

    :param game_object: Target Object.
    :param collection: Target Collection.
    :param exclusive: Remove the object from all other collections.
    """
    if exclusive:
        for coll in game_object.blenderObject.users_collection:
            coll.objects.unlink(game_object.blenderObject)
    if isinstance(collection, str):
        collection = bpy.data.collections.get(collection)
    collection.objects.link(game_object.blenderObject)
    game_object.blenderObject.update_tag()


class Collection:

    @property
    def game_objects(self):
        scene = logic.getCurrentScene()
        return [scene.getGameObjectFromObject(bobj) for bobj in self.collection.objects]

    @property
    def all_game_objects(self):
        scene = logic.getCurrentScene()
        objs = [scene.getGameObjectFromObject(bobj) for bobj in self.collection.all_objects]
        for o in objs:
            if o.groupMembers is not None and o.groupObject is None:
                objs.extend(o.groupMembers)
        return objs

    @property
    def objects(self):
        return self.collection.objects

    @property
    def all_objects(self):
        return self.collection.all_objects

    def __init__(self, collection: BColl) -> None:
        if isinstance(collection, str):
            collection = bpy.data.collections.get(collection)
        self.collection = collection
        if collection is None:
            return
        self._collection_state = {
            'objects': []
        }
        self.save()

    def set_frozen(self, state):
        for obj in self.all_game_objects:
            if not (obj.groupMembers is not None and obj.groupObject is None):
                obj.suspendPhysics() if state else obj.restorePhysics()
                obj.suspendDynamics() if state else obj.restoreDynamics()

    def set_visible(self, state=True, physics=True):
        for obj in self.all_game_objects:
            if obj.groupMembers is not None and obj.groupObject is None:
                # Is groupInstance
                obj.setVisible(False, True)
            else:
                obj.setVisible(state, False)
                if physics:
                    obj.restorePhysics() if state else obj.suspendPhysics()
                    obj.restoreDynamics() if state else obj.suspendDynamics()
        bpy.context.scene.update_tag()

    def enable(self):
        self.set_visible(True)

    def disable(self):
        self.set_visible(False)

    def get_game_vec(self, data):
        return Euler((data['x'], data['y'], data['z']))

    def load(self):
        scene = logic.getCurrentScene()
        for obj in self.all_objects:
            data = self._collection_state['objects'].get(obj.name, None)
            if data is None:
                continue
            game_obj = scene.getGameObjectFromObject(obj)

            lPos = self.get_game_vec(data['data']['localPosition'])
            lOri = self.get_game_vec(data['data']['localOrientation'])
            lSca = self.get_game_vec(data['data']['localScale'])

            wPos = self.get_game_vec(data['data']['worldPosition'])
            wOri = self.get_game_vec(data['data']['worldOrientation'])
            wSca = self.get_game_vec(data['data']['worldScale'])
            
            game_obj.worldPosition = wPos
            game_obj.worldOrientation = wOri.to_matrix()
            game_obj.worldScale = wSca

            if game_obj.parent:
                game_obj.localPosition = lPos
                game_obj.localOrientation = lOri.to_matrix()
                game_obj.localScale = lSca

            if data['type'] == 'rigid_body':
                linVel = self.get_game_vec(
                    data['data']['worldLinearVelocity']
                )
                angVel = self.get_game_vec(
                    data['data']['worldAngularVelocity']
                )
                game_obj.worldLinearVelocity = linVel
                game_obj.worldAngularVelocity = angVel

            if data['type'] == 'light':
                energy = data['data']['energy']
                game_obj.energy = energy

            if data['type'] == 'character':
                wDir = self.get_game_vec(data['data']['walkDirection'])
                (
                    constraints
                    .getCharacter(game_obj)
                    .walkDirection
                ) = wDir

            for prop in data['data']['props']:
                game_obj[prop['name']] = prop['value']

    def save(self, properties=True):
        self._collection_state = {
            'objects': {}
        }
        objs = self._collection_state['objects']

        for obj in self.all_game_objects:
            if obj.name == '__default__cam__':
                continue
            props = obj.getPropertyNames()
            prop_list = []
            cha = constraints.getCharacter(obj)
            if properties:
                for prop in props:
                    if prop.startswith('NL__'):
                        continue
                    if isinstance(obj[prop], Vector):
                        continue
                    prop_set = {}
                    prop_set['name'] = prop
                    prop_set['value'] = obj[prop]
                    prop_list.append(prop_set)
            
            locloc = obj.localPosition
            locrot = obj.localOrientation.to_euler()
            locsca = obj.localScale
            
            loc = obj.worldPosition
            rot = obj.worldOrientation.to_euler()
            sca = obj.worldScale

            if obj.blenderObject.game.physics_type == 'RIGID_BODY':
                lin_vel = obj.worldLinearVelocity
                ang_vel = obj.worldAngularVelocity
                loclin_vel = obj.localLinearVelocity
                locang_vel = obj.localAngularVelocity

                objs[obj.blenderObject.name] = {
                        'name': obj.name,
                        'type': 'rigid_body',
                        'data': {
                            'localPosition': {
                                'x': locloc.x,
                                'y': locloc.y,
                                'z': locloc.z
                            },
                            'localOrientation': {
                                'x': locrot.x,
                                'y': locrot.y,
                                'z': locrot.z
                            },
                            'localScale': {
                                'x': locsca.x,
                                'y': locsca.y,
                                'z': locsca.z
                            },
                            'worldPosition': {
                                'x': loc.x,
                                'y': loc.y,
                                'z': loc.z
                            },
                            'worldOrientation': {
                                'x': rot.x,
                                'y': rot.y,
                                'z': rot.z
                            },
                            'worldLinearVelocity': {
                                'x': lin_vel.x,
                                'y': lin_vel.y,
                                'z': lin_vel.z
                            },
                            'worldAngularVelocity': {
                                'x': ang_vel.x,
                                'y': ang_vel.y,
                                'z': ang_vel.z
                            },
                            'localLinearVelocity': {
                                'x': loclin_vel.x,
                                'y': loclin_vel.y,
                                'z': loclin_vel.z
                            },
                            'localAngularVelocity': {
                                'x': locang_vel.x,
                                'y': locang_vel.y,
                                'z': locang_vel.z
                            },
                            'worldScale': {'x': sca.x, 'y': sca.y, 'z': sca.z},
                            'props': prop_list
                        }
                    }
            elif cha:
                wDir = cha.walkDirection

                objs[obj.blenderObject.name] = {
                        'name': obj.name,
                        'type': 'character',
                        'data': {
                            'localPosition': {
                                'x': locloc.x,
                                'y': locloc.y,
                                'z': locloc.z
                            },
                            'localOrientation': {
                                'x': locrot.x,
                                'y': locrot.y,
                                'z': locrot.z
                            },
                            'localScale': {
                                'x': locsca.x,
                                'y': locsca.y,
                                'z': locsca.z
                            },
                            'worldPosition': {
                                'x': loc.x,
                                'y': loc.y,
                                'z': loc.z
                            },
                            'worldOrientation': {
                                'x': rot.x,
                                'y': rot.y,
                                'z': rot.z
                            },
                            'worldScale': {'x': sca.x, 'y': sca.y, 'z': sca.z},
                            'walkDirection': {
                                'x': wDir.x,
                                'y': wDir.y,
                                'z': wDir.z
                            },
                            'props': prop_list
                        }
                    }
            else:
                objs[obj.blenderObject.name] = {
                        'name': obj.name,
                        'type': 'static',
                        'data': {
                            'localPosition': {
                                'x': locloc.x,
                                'y': locloc.y,
                                'z': locloc.z
                            },
                            'localOrientation': {
                                'x': locrot.x,
                                'y': locrot.y,
                                'z': locrot.z
                            },
                            'localScale': {
                                'x': locsca.x,
                                'y': locsca.y,
                                'z': locsca.z
                            },
                            'worldPosition': {
                                'x': loc.x,
                                'y': loc.y,
                                'z': loc.z
                            },
                            'worldOrientation': {
                                'x': rot.x,
                                'y': rot.y,
                                'z': rot.z
                            },
                            'worldScale': {'x': sca.x, 'y': sca.y, 'z': sca.z},
                            'props': prop_list
                        }
                    }
            self._collection_state['objects'] = objs
            



from uplogic.nodes import ULActionNode
from uplogic.audio.fmod import FMod
import os


class FModLoadBankNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = False
        self.load_master = False
        self.load_strings = False
        self.path = ''
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        path = self.get_input(self.path)
        if self.load_master:
            ppath = os.path.join(*os.path.split(path)[:-1])
            FMod.load_bank(os.path.join(ppath, 'Master.bank'))
            if self.load_strings:
                FMod.load_bank(os.path.join(ppath, 'Master.strings.bank'))
        FMod.load_bank(self.get_input(self.path))
        self._done = True



from bge.types import KX_GameObject
from gpu.types import GPUOffScreen
from gpu_extras.presets import draw_texture_2d
import bge
import bpy
import gpu


class RenderToScreen:

    def __init__(self, camera: KX_GameObject, resolution=(1920, 1080), pos=(0, 0)) -> None:
        self.offscreen = gpu.types.GPUOffScreen(resolution[0], resolution[1])
        self.resolution = resolution  # [int(resolution[0] * .1), int(resolution[1] * 0.1)]
        self.pos = pos

        context = bpy.context
        context.space_data.overlay.show_overlays = False
        bpy.app.handlers.game_post.append(self.stop)
        bge.logic.getCurrentScene().post_draw.append(self.draw)
        self.scene = bpy.data.scenes.get(camera.scene.name, None)
        if self.scene is None:
            return
        self.camera = camera.blenderObject
        self.shader = GPUOffScreen(int(resolution[0] * .01), int(resolution[1] * 0.01))

    def stop(self, *a):
        context = bpy.context
        context.space_data.overlay.show_overlays = True

    def draw(self):
        context = bpy.context
        scene = self.scene
        view_matrix = self.camera.matrix_world.inverted()
        width, height = self.resolution
        projection_matrix = self.camera.calc_matrix_camera(
            context.evaluated_depsgraph_get(),
            x=width,
            y=height
        )

        self.offscreen.draw_view3d(
            scene,
            context.view_layer,
            context.space_data,
            context.region,
            view_matrix,
            projection_matrix,
            do_color_management=True
        )

        gpu.state.depth_mask_set(False)
        draw_texture_2d(self.offscreen.texture_color, self.pos, width, height)



from uplogic.nodes import ULParameterNode


class ULGetPyInstanceAttr(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.instance = None
        self.attr = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        instance = self.get_input(self.instance)
        attr = self.get_input(self.attr)
        return getattr(instance, attr, None)



from bge import constraints
from uplogic.nodes import ULActionNode


class ULRemovePhysicsConstraint(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.object = None
        self.name = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        obj = self.get_input(self.object)
        name = self.get_input(self.name)
        constraints.removeConstraint(obj[name].getConstraintId())
        self._done = True



'''TODO: Documentation
'''
from .lights import ULLight, Light  # noqa
from .lights import make_unique_light  # noqa
from .nodetrees import get_geom_socket  # noqa
from .nodetrees import get_group_socket  # noqa
from .nodetrees import get_material_socket  # noqa
from .nodetrees import get_world_socket  # noqa
from .nodetrees import modify_geom_socket  # noqa
from .nodetrees import modify_material_socket  # noqa
from .nodetrees import modify_world_socket  # noqa
from .nodetrees import set_geom_socket  # noqa
from .nodetrees import set_group_socket  # noqa
from .nodetrees import set_material_socket  # noqa
from .nodetrees import set_world_socket  # noqa
from .objects import get_curve_length
from .objects import ULCurve, Curve  # noqa
from .objects import Mesh  # noqa
from .objects import controller_brick_status  # noqa
from .objects import controller_brick  # noqa
from .objects import create_curve  # noqa
from .objects import set_curve_points  # noqa
from .raycasting import raycast  # noqa
from .raycasting import raycast_screen  # noqa
from .raycasting import raycast_camera  # noqa
from .raycasting import raycast_face  # noqa
from .raycasting import raycast_projectile  # noqa
from .raycasting import raycast_mouse  # noqa
from .scene import set_scene  # noqa
from .scene import get_custom_loop  # noqa
from .pooling import SpawnPool  # noqa
from .pooling import Spawn  # noqa
from .pooling import SimpleBullet  # noqa
from .pooling import PhysicsBullet  # noqa
from .visualize import draw_box  # noqa
from .visualize import draw_cube  # noqa
from .visualize import draw_line  # noqa
from .visualize import draw_path  # noqa
from .visualize import draw_mesh  # noqa
from .visualize import draw_arrow  # noqa
from .visualize import draw_arrow_path  # noqa
from .visualize import draw_axis  # noqa
from .scene import FileLoader  # noqa
from .scene import SceneLoader  # noqa
from .math import clamp
from .math import cycle
from .math import vec_abs
from .math import vec_clamp
from .math import interpolate
from .math import lerp
from .math import get_angle
from .math import get_bitmask
from .math import get_collision_bitmask
from .math import get_direction
from .math import get_local
from .math import get_raw_angle
from .math import map_range
from .scene import screen_to_world
from .scene import world_to_screen
from .math import rotate2d
from .math import rotate3d
from .math import rotate_by_axis
from .math import rotate_by_euler
from .objects import xrot_to
from .objects import yrot_to
from .objects import zrot_to
from .objects import rotate_to
from .constants import WATER
from .constants import OPERATORS
from .constants import MATH_OPERATORS
from .constants import LOGIC_OPERATORS
from .constants import RED
from .constants import GREEN
from .constants import BLUE
from .constants import YELLOW
from .constants import PURPLE
from .constants import TORQUISE
from .constants import WHITE
from .constants import BLACK
from .constants import GREY
from .constants import FPS_FACTOR
from .constants import DELTA_TIME
from bge import logic
from bge.types import KX_GameObject as GameObject
from mathutils import Matrix
from mathutils import Vector
import time as t
from os import path, pardir

import bpy
import json
import math


###############################################################################
# LOGIC NODES
###############################################################################


def _name_query(named_items, query):
    assert len(query) > 0
    postfix = (query[0] == "*")
    prefix = (query[-1] == "*")
    infix = (prefix and postfix)
    if infix:
        token = query[1:-1]
        for item in named_items:
            if token in item.name:
                return item
    if prefix:
        token = query[:-1]
        for item in named_items:
            if item.name.startswith(token):
                return item
    if postfix:
        token = query[1:]
        for item in named_items:
            if item.name.endswith(token):
                return item
    for item in named_items:
        if item.name == query:
            return item
    return None


def check_game_object(query, scene=None):
    '''TODO: Documentation
    '''
    if not scene:
        scene = logic.getCurrentScene()
    else:
        scene = scene
    if (query is None) or (query == ""):
        return
    if not is_invalid(scene):
        # find from scene
        return _name_query(scene.objects, query)


def compute_distance(parama, paramb) -> float:
    '''TODO: Documentation
    '''
    if is_invalid(parama):
        return None
    if is_invalid(paramb):
        return None
    if hasattr(parama, "getDistanceTo"):
        return parama.getDistanceTo(paramb)
    if hasattr(paramb, "getDistanceTo"):
        return paramb.getDistanceTo(parama)
    va = Vector(parama)
    vb = Vector(paramb)
    return (va - vb).length


# def debug(message: str):
#     if not hasattr(bpy.types.Scene, 'logic_node_settings'):
#         return
#     if not bpy.context or not bpy.context.scene:
#         return
#     if not bpy.context.scene.logic_node_settings.use_node_debug:
#         return
#     else:
#         print('[UPLOGIC] ' + message)


def is_invalid(*a) -> bool:
    for ref in a:
        if ref is None or ref == '':
            return True
        if not hasattr(ref, "invalid"):
            continue
        elif ref.invalid:
            return True
    return False


def make_valid_name(name):
    valid_characters = (
        "_abcdefghijklmnopqrstuvwxyz1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    )
    clsname = name.replace(' ', '_')
    stripped_name = "".join(
        [c for c in clsname if c in valid_characters]
    )
    return stripped_name


def not_met(*conditions) -> bool:
    for c in conditions:
        if (
            c is None or
            c is False
        ):
            return True
    return False


def load_user_module(module_name):
    import sys
    exec(f"import {module_name}")
    return sys.modules[module_name]


def unload_nodes(a, b):
    if not hasattr(bpy.types.Scene, 'nl_globals_initialized'):
        return
    delattr(bpy.types.Scene, 'nl_globals_initialized')


###############################################################################
# SCENE
###############################################################################


def get_closest_instance(game_obj: GameObject, name: str):
    '''TODO: Documentation
    '''
    objs = []
    distances = {}
    for obj in logic.getCurrentScene().objects:
        if obj.name == name:
            objs.append(obj)
    for obj in objs:
        distances[game_obj.getDistanceTo(obj)] = obj
    return distances[min(distances.keys())]


def is_water(game_object: GameObject):
    return WATER in game_object.getPropertyNames()


def get_child_by_name(obj: GameObject, child: str, recursive: bool = True, partial: bool = False) -> GameObject:
    """Get a named child object.

    :param obj: The parent object.
    :param child: The child's name.
    :param recursive: If True, children's children will be seached too.
    :param partial: If a child object name has the given name in it, it counts as a hit.
    """
    children = obj.childrenRecursive if recursive else obj.children
    if partial:
        for c in children:
            if child in c.name:
                return c
    else:
        return children.get(child)


def check_vr_session_status() -> bool:
    """Check if a VR session is currently running.
    """
    session = bpy.context.window_manager.xr_session_state
    return session is not None


def get_project_folder(folder_name, *structure):
    from uplogic.console import warning
    warning("'get_project_folder' is deprecated, please use 'get_project_path' instead.")
    return get_project_path(folder_name, *structure)


def get_project_path(folder_name, *structure):
    from uplogic.console import error
    directory = og_path = path.join(bpy.path.abspath('//'))[:-1]
    while not directory.endswith(folder_name):
        # print(folder_name)
        pdir = path.abspath(path.join(directory, pardir))
        if pdir == directory:
            error(f"Can't go beyond drive's root from {og_path}!")
            return ''
        directory = pdir
    return path.join(directory, *structure)


from uplogic.nodes import ULParameterNode
from mathutils import Vector
from uplogic.utils.math import angle_signed
from uplogic.utils.math import wrap
from uplogic.utils.math import snap
import math


class VectorMathNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.operator = 0
        self.vector_a = None
        self.vector_b = None
        self.vector_c = None
        self.value = None
        self.RESULT = self.add_output(self.get_done)
        self.RESULT_VECTOR = self.add_output(self.get_done)
        
        self.calculations = [
            self.get_add,  # 0
            self.get_subtract,  # 1
            self.get_multiply,  # 2
            self.get_divide,  # 3
            self.get_multadd,  # 4
            self.get_matmul,  # 5
            self.get_angle,  # 6
            self.get_angle_signed,  # 7
            self.get_cross,  # 8
            self.get_project,  # 9
            self.get_reflect,  # 10
            self.get_refract,  # 11
            self.get_faceforward,  # 12
            self.get_dot,  # 13
            self.get_lerp,  # 14
            self.get_slerp,  # 15
            self.get_distance,  # 16
            self.get_length,  # 17
            self.get_scale,  # 18
            self.get_normalize,  # 19
            self.get_absolute,  # 20
            self.get_negate,  # 21
            self.get_min,  # 22
            self.get_max,  # 23
            self.get_round,  # 24
            self.get_floor,  # 25
            self.get_ceil,  # 26
            self.get_fraction,  # 27
            self.get_mod,  # 28
            self.get_wrap,  # 29
            self.get_snap,  # 30
            self.get_sine,  # 31
            self.get_cosine,  # 32
            self.get_tangent  # 33
        ]

    def get_done(self):
        vector_a = self.get_input(self.vector_a)
        vector_b = self.get_input(self.vector_b)
        vector_c = self.get_input(self.vector_c)
        value = self.get_input(self.value)
        if vector_a:
            vector_a = vector_a.copy()
        if vector_b:
            vector_b = vector_b.copy()
        if vector_c:
            vector_c = vector_c.copy()
        return self.calculations[self.operator](vector_a, vector_b, vector_c, value)

    def get_scale(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a * value

    def get_length(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a.length

    def get_distance(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return (vector_a - vector_b).length

    def get_angle(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a.angle(vector_b)

    def get_angle_signed(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return angle_signed(vector_a, vector_b, vector_c)

    def get_dot(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a.dot(vector_b)

    def get_faceforward(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a if vector_c.dot(vector_b) < 0.0 else -vector_a

    def get_refract(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        k = 1.0 - value * value * (1.0 - vector_b.dot(vector_a) * vector_b.dot(vector_a))
        if (k < 0):
            return Vector((0, 0, 0))
        else:
            return value * vector_a - (value * vector_b.dot(vector_a) + math.sqrt(k)) * vector_b

    def get_reflect(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a.reflect(vector_b)

    def get_project(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a.project(vector_b)

    def get_cross(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a.cross(vector_b)

    def get_multadd(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return (vector_a * vector_b) + vector_c

    def get_divide(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([e / vector_b[i] if vector_b[i] != 0 else 0 for i, e in enumerate(vector_a)])

    def get_multiply(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a * vector_b

    def get_matmul(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a @ vector_b

    def get_subtract(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a - vector_b

    def get_add(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a + vector_b

    def get_normalize(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a.normalized()

    def get_lerp(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a.lerp(vector_b, value)

    def get_slerp(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return vector_a.slerp(vector_b, value, vector_c)

    def get_negate(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        vector_a.negate()
        return vector_a

    def get_absolute(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([abs(e) for e in vector_a])

    def get_min(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([min(e) for e in vector_a])

    def get_max(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([max(e) for e in vector_a])

    def get_round(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([round(e) for e in vector_a])

    def get_floor(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([math.floor(e) for e in vector_a])

    def get_ceil(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([math.ceil(e) for e in vector_a])

    def get_fraction(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([e - math.floor(e) for e in vector_a])

    def get_mod(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([e % vector_b[i] if vector_b[i] != 0 else 0 for i, e in enumerate(vector_a)])

    def get_wrap(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([wrap(e, vector_b[i]) for i, e in enumerate(vector_a)])

    def get_snap(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([snap(e, vector_b[i]) for i, e in enumerate(vector_a)])

    def get_sine(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([e - math.sin(e) for e in vector_a])

    def get_cosine(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([e - math.cos(e) for e in vector_a])

    def get_tangent(self, vector_a: Vector, vector_b: Vector, vector_c: Vector, value: float):
        return Vector([e - math.tan(e) for e in vector_a])



from bge import events
from bge import logic
from bge.types import KX_GameObject as GameObject
from uplogic.audio import AudioSystem
from uplogic.data import GlobalDB
from uplogic.nodes import ULLogicContainer
from uplogic.utils import load_user_module
from uplogic.utils import make_valid_name
from uplogic.data import init_glob_cats
import bpy
# from uplogic import get_mainloop
from bge.types import SCA_PythonKeyboard as Keyboard
import collections
from bge.logic import getRealTime
from uplogic import console


class ULLogicTree(ULLogicContainer):

    def __init__(self):
        from ..input import Mouse
        ULLogicContainer.__init__(self)
        self._cells: list = []
        self._iter = collections.deque()
        self._lastuid: int = 0
        self.component = {}
        self._owner: GameObject = None
        self._initialized = False
        self._max_blocking_loop_count: int = 0
        self.keyboard: Keyboard = None
        self.mouse: Mouse = Mouse()
        self.mouse_events = None
        self.stopped = False
        self._time_then = getRealTime()
        self.time_per_frame = 0.0
        self._do_remove = False
        self.aud_system_owner = False
        init_glob_cats()
        self.audio_system = self.get_aud_system()
        self.sub_networks = []  # a list of networks updated by this network
        self.capslock_pressed = False
        self.evaluated_cells = 0
        # scene = self.scene = logic.getCurrentScene()
        mainloop = logic.globalDict.get('loop')
        if mainloop:
            mainloop.logic_tree = self

    @property
    def owner(self):
        return self._owner

    def get_aud_system(self):
        aud_sys = GlobalDB.retrieve('uplogic.audio').get('default')
        if not aud_sys:
            self.aud_system_owner = True
            return AudioSystem('default')
        return aud_sys

    def set_mouse_position(self, screen_x, screen_y):
        self.mouse.position = (screen_x, screen_y)

    def get_owner(self):
        return self._owner

    def setup(self):
        self.time_per_frame = 0.0
        for cell in self._cells:
            cell.network = self
            cell.setup(self)

    def is_running(self):
        return not self.stopped

    def is_stopped(self):
        return self.stopped

    def stop(self, network=None):
        if self.stopped:
            return
        # self._time_then = None  # XXX: Was this important? Causes issues when stopping and restarting
        self.stopped = True
        for cell in self._cells:
            cell.stop(self)

    def _generate_cell_uid(self):
        self._lastuid += 1
        return self._lastuid

    def add_cell(self, cell):
        self._cells.append(cell)
        self._iter.append(cell)
        self._max_blocking_loop_count = len(self._cells) * len(self._cells)
        cell._uid = self._generate_cell_uid()
        return cell

    def evaluate(self):
        now = getRealTime()
        dtime = now - self._time_then
        self._time_then = now
        self.time_per_frame = dtime
        if self._owner.invalid:
            console.debug("Network Owner removed from game. Shutting down the network")
            return True
        self.keyboard = logic.keyboard
        self.keyboard_events = self.keyboard.inputs.copy()
        caps_lock_event = self.keyboard_events[events.CAPSLOCKKEY]
        if(caps_lock_event.released):
            self.capslock_pressed = not self.capslock_pressed
        # update the cells
        cells = self._iter.copy()
        while cells:
            cell = cells.popleft()
            cell.evaluate()
        for cell in self._cells:
            cell.reset()
        self._initialized = True
        for network in self.sub_networks:
            if network._owner.invalid:
                self.sub_networks.remove(network)
            elif not network.stopped:
                network.evaluate()

    def install_subnetwork(self, owner_object, node_tree_name, initial_status):
        # transform the tree name into a NL module name
        tree_name = make_valid_name(node_tree_name)
        mname = f'nl_{tree_name.lower()}'
        module = load_user_module(mname)
        owner_object[f'NL__{tree_name}'] = initial_status
        tree = module.get_tree(owner_object).network
        owner_object[f'IGNLTree_{tree_name}'] = tree
        self.sub_networks.append(tree)
        tree.stopped = not initial_status
        return tree



from .widget import Widget
import gpu
import bge, bpy
from gpu_extras.batch import batch_for_shader
from gpu_extras.presets import draw_texture_2d
from mathutils import Vector
from bge import render
from uplogic.shaders import Buffer


class RenderedTexture(Widget):

    def __init__(self):
        super().__init__()
        self.buffer = Buffer()
        self.start()

    def _build_shader(self):
        size = self._draw_size
        self.view = gpu.types.GPUOffScreen(size[0], size[1])

        pos = self._draw_pos
        size = self._draw_size
        x0 = Vector((pos[0], pos[1]))
        x1 = Vector((pos[0] + size[0], pos[1]))
        y1 = Vector((pos[0] + size[0], pos[1] + size[1]))
        y0 = Vector((pos[0], pos[1] + size[1]))

        vertices = self._vertices = (
            x1, x0, y1, y0
        )

        tex_vert_shader = """
        in vec2 texCoord;
        in vec2 pos;
        out vec2 uv;

        uniform mat4 ModelViewProjectionMatrix;

        void main() {
            uv = texCoord;
            gl_Position = ModelViewProjectionMatrix * vec4(pos.xy, 0.0, 1.0);
        }
        """

        tex_frag_shader = """
        in vec2 uv;
        out vec4 fragColor;

        uniform sampler2D renderedTexture;

        void main() {
            vec4 image = texture(renderedTexture, uv);
            fragColor = image;
        }
        """
        self._shader = gpu.types.GPUShader(tex_vert_shader, tex_frag_shader)
        self._batch = batch_for_shader(
            self._shader, 'TRI_STRIP',
            {
                "pos": vertices,
                "texCoord": (
                    (.9999, .0001),
                    (0.0001, 0.0001),
                    (.9999, .9999),
                    (.0001, .9999)
                ),
            },
        )


    def draw(self):
        self._shader.uniform_sampler("renderedTexture", self.buffer.texture)
        self._setup_draw()
        self._batch.draw(self._shader)
        super().draw()



from uplogic.nodes import ULActionNode


class ULSetMousePosition(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.screen_x = None
        self.screen_y = None
        self.network = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def setup(self, network):
        self.network = network

    def evaluate(self):
        if not self.get_condition():
            return
        screen_x = self.get_input(self.screen_x)
        screen_y = self.get_input(self.screen_y)
        self.network.set_mouse_position(screen_x, screen_y)
        self._done = True



from uplogic.nodes import ULParameterNode


class ULGetActuatorValue(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_obj = None
        self.act_name = None
        self.field = None
        self.OUT = self.add_output(self.get_actuator)

    def get_actuator(self):
        game_obj = self.get_input(self.game_obj)
        act_name = self.get_input(self.act_name)
        field = self.get_input(self.field)
        if act_name not in game_obj.actuators:
            return False
        return getattr(game_obj.actuators[act_name], field)



from uplogic.nodes import ULParameterNode
from uplogic.utils import Curve
from mathutils import Vector


class EvaluateCurveNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.curve = None
        self.factor = 0.0
        self._curve = None
        self._factor = None
        self._vec = Vector((0, 0, 0))
        self.VEC = self.add_output(self.get_out_v)

    def get_out_v(self):
        recalc = False
        curve = self.get_input(self.curve)
        if self._curve is None or curve is not self._curve.game_object:
            self._curve = Curve(curve, use_evaluate=True)
            recalc = True
        fac = self.get_input(self.factor)
        if fac != self._factor:
            self._factor = fac
            recalc = True
        if recalc:
            self._vec = self._curve.evaluate(self._factor)
        return self._vec



from uplogic.nodes import ULConditionNode


class ULValueChangedTo(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.monitored_value = None
        self.trigger_value = None
        self.changed = None
        self.old = None
        self.OUT = self.add_output(self.get_changed)

    def get_changed(self):
        monitored_value = self.get_input(self.monitored_value)
        trigger_value = self.get_input(self.trigger_value)
        return (
            monitored_value == trigger_value and
            self.changed
        )

    def evaluate(self):
        monitored_value = self.get_input(self.monitored_value)
        self.changed = monitored_value != self.old
        self.old = monitored_value



from uplogic.nodes import ULActionNode
from uplogic.utils import make_valid_name


class ULStopSubNetwork(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.logic_network_name = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        logic_network_name = self.get_input(self.logic_network_name)
        tree_name = make_valid_name(logic_network_name)
        network = game_object.get(f'IGNLTree_{tree_name}')
        if network is None:
            return
        network.stop()
        self._done = True



from mathutils import Vector
from uplogic.nodes import ULParameterNode


class ULObjectAttribute(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.attribute_name = ''
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        game_object = self.get_input(self.game_object)
        attribute_name = self.get_input(self.attribute_name)
        if not hasattr(game_object, attribute_name):
            return Vector((0, 0, 0))
        val = getattr(game_object, attribute_name, None)
        return val.copy() if isinstance(val, Vector) else val



from uplogic.nodes import ULActionNode
from bge.types import KX_GameObject
from bpy.types import Mesh


class ULReplaceMesh(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.target_game_object = None
        self.new_mesh_name = None
        self.use_display = None
        self.use_physics = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        target: KX_GameObject = self.get_input(self.target_game_object)
        mesh: Mesh = self.get_input(self.new_mesh_name)
        display: bool = self.get_input(self.use_display)
        physics: bool = self.get_input(self.use_physics)
        target.replaceMesh(mesh.name, display, physics)
        if physics:
            target.reinstancePhysicsMesh()
        self._done = True



from uplogic.nodes import ULConditionNode
from bge.logic import getRealTime


class ULBarrier(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = None
        self.time = None
        self.consumed = False
        self.trigger = 0
        self.result = False
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self.result

    def evaluate(self):
        self.result = False
        time = self.get_input(self.time)
        now = getRealTime()
        if self.get_input(self.condition):
            if not self.consumed:
                self.consumed = True
                self.trigger = now + time

            if now >= self.trigger:
                self.result = True

        else:
            self.result = False
            self.consumed = False



from uplogic.nodes import ULActionNode
from uplogic.utils import clamp


class ULClampedModifyProperty(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.property_name = None
        self.property_value = None
        self.mode = 0
        self.operator = None
        self.range = None
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        property_name = self.get_input(self.property_name)
        property_value = self.get_input(self.property_value)
        val_range = self.get_input(self.range)
        obj = game_object.blenderObject if self.mode else game_object
        value = obj.get(property_name, 0)
        new_val = self.operator(value, property_value)
        obj[property_name] = (
            clamp(new_val, val_range.x, val_range.y)
        )
        self._done = True



from uplogic.nodes import ULParameterNode
from uplogic import console
import bpy
import json
import os


class ULLoadVariable(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.name = None
        self.default_value = None
        self.file_name = None
        self.path = ''
        self.VAR = self.add_output(self.get_var)

    def get_var(self):
        name = self.get_input(self.name)
        path = self.get_input(self.path)
        path = (bpy.path.abspath(path))

        os.makedirs(path, exist_ok=True)

        return self.read_from_json(path, name)

    def read_from_json(self, path, name):
        if not path.endswith('.json'):
            path = os.path.join(path, f'{self.get_input(self.file_name)}.json')
        if os.path.isfile(path):
            f = open(path, 'r')
            data = json.load(f)
            if name not in data:
                console.debug(f'"{name}" is not a saved Variabe!')
                return self.get_input(self.default_value)
            f.close()
            return data.get(name)
        else:
            console.debug('No saved variables!')
            return self.get_input(self.default_value)





from uplogic.nodes import ULConditionNode
from uplogic.input import gamepad_tap
from uplogic.input import gamepad_down
from uplogic.input import gamepad_up


class ULGamepadButton(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False
        self.input_type = 1 if self.pulse else 0
        self.button = 0
        self.index = None
        self._button = None
        self.BUTTON = self.add_output(self.get_button)
        self.initialized = False

    def get_button(self):
        index = self.get_input(self.index)
        funcs = [gamepad_tap, gamepad_down, gamepad_up]
        return funcs[self.input_type](self.button, index)



from ..widget import Widget
from .canvas import Canvas
from .layout import FloatLayout, RelativeLayout, GridLayout, BoxLayout, PolarLayout
import math
from mathutils import Vector

def rotate2d(origin, pivot, angle):
    angle = math.radians(angle)
    return Vector((
        ((origin[0] - pivot[0]) * math.cos(angle)) - ((origin[1] - pivot[1]) * math.sin(angle)) + pivot[0],
        ((origin[0] - pivot[0]) * math.sin(angle)) + ((origin[1] - pivot[1]) * math.cos(angle)) + pivot[1]
    ))


from uplogic.nodes import ULParameterNode
from uplogic.utils.portals import Portals


class GetPortalNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.VAL = self.add_output(self.get_value)
        self._portal = None
    
    @property
    def portal(self):
        return self._portal

    @portal.setter
    def portal(self, val):
        self._portal = Portals.get(val)

    def get_value(self):
        node = self.portal.node
        if node:
            return node.get_input(node.value)
        return None



from uplogic.nodes import ULParameterNode
from uplogic.animation.rig import Rig, RigBone


class GetRigBoneAttributeNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.armature = None
        self.bone = None
        self.world_space = False
        self.VALUE = self.add_output(self.get_val)
        self.attribute = 'inherit_rotation'
        self._rig: Rig = None

    def get_val(self):
        armature = self.get_input(self.armature)
        if self._rig is None or self._rig.game_object is not armature:
            self._rig = Rig(armature)

        bone: RigBone = self._rig.bones[self.get_input(self.bone)]
        self._rig.visualize()
        bone.world_space = self.world_space
        return getattr(bone, self.attribute)



from uplogic.data import GlobalDB
from uplogic.nodes import ULActionNode


class ULSetGlobalValue(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.data_id = None
        self.key = None
        self.value = None
        self.persistent = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        data_id = self.get_input(self.data_id)
        persistent = self.get_input(self.persistent)
        key = self.get_input(self.key)
        value = self.get_input(self.value)
        db = GlobalDB.retrieve(data_id)
        db.put(key, value, persistent)
        self._done = True



from uplogic.events import send
from uplogic.nodes import ULActionNode


class ULDispatchEvent(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.subject = None
        self.body = None
        self.messenger = None
        self.target = None
        self.old_subject = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        send(
            self.get_input(self.subject),
            self.get_input(self.body),
            self.get_input(self.messenger),
            self.get_input(self.target)
        )
        self._done = True



import socket
from pythonosc.dispatcher import Dispatcher
from pythonosc import osc_server
import threading
from uplogic import events
from uplogic import console


class OSC_Server:

    def __init__(self, ip=None, port=8304):
        self._mapped_callbacks = {}

        if ip is None:
            ip = socket.gethostbyname(socket.gethostname())

        self.dispatcher = Dispatcher()
        self.map('/icom', self._print_debug)
        self.server = osc_server.ThreadingOSCUDPServer(
            (ip, port), self.dispatcher)
        self.thread = threading.Thread(target=self.server.serve_forever)
        self.thread.start()
        console.debug("Serving on {}".format(self.server.server_address))

    def _print_debug(self, address, args):
        console.debug(address, args)

    def map(self, address: str, callback, *args: list, needs_reply_address: bool = False):
        """Map a callback to an address. Callback receives (`address`, `args`, `volume`)
        
        :param address: address identifier (example: `"/update"`).
        :param callback: Callback function that will be called as the handler for the given address.
        :param args: Fixed arguements that will be passed to the callback function.
        :param needs_reply_address: Whether the IP address from which the message originated from shall be passed as
        an argument to the handler callback identifier.
        """
        def _deferred_cb(*a):
            events.schedule(callback, content=a)

        handler = self.dispatcher.map(address, _deferred_cb, *args, needs_reply_address=needs_reply_address)
        handlers = self._mapped_callbacks.get(address, [])
        handlers.append(handler)
        self._mapped_callbacks[address] = handlers
        return handler

    def unmap(self, address, handler):
        try:
            self.dispatcher.unmap(address, handler)
            self._mapped_callbacks.get(address, []).remove(handler)
        except ValueError as e:
            console.error(e)

    def unmap_all(self):
        for address in self._mapped_callbacks.keys():
            for handler in self._mapped_callbacks.get(address, []).copy():
                self.unmap(address, handler)
        self._mapped_callbacks.clear()

    def shutdown(self) -> None:
        console.debug(f'Shutting down OSC Server on {self.server.server_address[0]}:{self.server.server_address[1]}')
        self.server.shutdown()
        self.server.socket.close()



from bge import logic
from uplogic.nodes import ULActionNode


class ULLoadBlendFile(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.file_name = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        file_name = self.get_input(self.file_name)
        logic.startGame(file_name)
        self._done = True



from bge import logic
from uplogic.audio import Sound3D
from uplogic.nodes import ULActionNode


class ULStartSpeaker(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.occlusion = None
        self.transition = None
        self.cutoff = None
        self.loop_count = None
        self.ignore_timescale = None
        self.speaker = None
        self.on_finish = False
        self._clear_sound = 1
        self._sustained = 1
        self._handle = None
        self.DONE = self.add_output(self.get_done)
        self.ON_FINISH = self.add_output(self.get_on_finish)
        self.HANDLE = self.add_output(self.get_handle)

    def get_handle(self):
        return self._handle

    def get_on_finish(self):
        if not self._handle:
            return False
        if self._handle.finished:
            self._handle = None
            return True
        return False

    def get_done(self):
        return self._done

    def evaluate(self):
        self.on_finish = False
        if not self.get_condition():
            return
        speaker = self.get_input(self.speaker)
        transition = self.get_input(self.transition)
        occlusion = self.get_input(self.occlusion)
        cutoff = self.get_input(self.cutoff)
        loop_count = self.get_input(self.loop_count)
        ignore_timescale = self.get_input(self.ignore_timescale)
        bl_speaker = speaker.blenderObject.data
        file = bl_speaker.sound.filepath
        if not file:
            return
        volume = bl_speaker.volume
        pitch = bl_speaker.pitch * logic.getTimeScale()
        attenuation = bl_speaker.attenuation
        distance_ref = bl_speaker.distance_reference
        cone_inner = bl_speaker.cone_angle_inner
        cone_outer = bl_speaker.cone_angle_outer
        cone_outer_volume = bl_speaker.cone_volume_outer
        self._handle = Sound3D(
            speaker,
            file,
            occlusion,
            transition,
            cutoff,
            loop_count,
            pitch,
            volume,
            False,
            attenuation,
            distance_ref,
            [cone_inner, cone_outer],
            cone_outer_volume,
            ignore_timescale,
            'ln_audio_system'
        )
        self._done = True



from uplogic.nodes import ULParameterNode
from mathutils import Vector


class ULWithinRange(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.range = None
        self.min_value = None
        self.max_value = None
        self.operator = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        v = self.get_input(self.value)
        if self.min_value is not None:
            r = Vector((self.get_input(self.min_value), self.get_input(self.max_value)))
        else:
            r = self.get_input(self.range)
        value = self.calc_range(self.operator, v, r)
        return value

    def calc_range(self, op, v, r):
        if op == 'OUTSIDE':
            return True if (v < r.x or v > r.y) else False
        if op == 'INSIDE':
            return True if (r.x < v < r.y) else False



from bge import logic
from uplogic.nodes import ULParameterNode


class ULGetScene(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return logic.getCurrentScene()



from uplogic.nodes import ULConditionNode


class ULOnValueChanged(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.old = None
        self.new = None
        self.current_value = None
        self.initialize = False
        self.OUT = self.add_output(self.get_changed)
        self.OLD = self.add_output(
            self.get_previous_value
        )
        self.NEW = self.add_output(self.get_current_value)

    def get_changed(self):
        curr = self.get_input(self.current_value)
        if not self.initialize:
            self.initialize = True
            self.old = self.new = curr
            return False
        elif self.old != curr:
            self.new = curr
            return True

    def get_previous_value(self):
        return self.old

    def get_current_value(self):
        return self.new

    def reset(self):
        super().reset()
        self.old = self.new



from uplogic.nodes import ULParameterNode
from random import randint
from random import random
from random import uniform
from mathutils import Vector


class ULRandomValue(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.min_float = None
        self.max_float = None
        self.min_int = None
        self.max_int = None
        self.min_vector = None
        self.max_vector = None
        self.probability = None
        self.seed = None
        self.OUT = self.add_output(self.get_float)

        self.getters = [
            self.get_float,
            self.get_integer,
            self.get_vector,
            self.get_bool
        ]

        self.data_type = 0

    @property
    def data_type(self):
        return None

    @data_type.setter
    def data_type(self, val):
        self.OUT._value_getter = self.getters[val]

    def get_float(self):
        return uniform(
            self.get_input(self.min_float),
            self.get_input(self.max_float)
        )

    def get_integer(self):
        return randint(
            self.get_input(self.min_int),
            self.get_input(self.max_int)
        )

    def get_vector(self):
        vec1 = self.get_input(self.min_vector)
        vec2 = self.get_input(self.max_vector)
        if not(isinstance(vec1, Vector) and isinstance(vec2, Vector)):
            return Vector((0, 0, 0))
        vec = Vector((
            uniform(vec1.x, vec2.x),
            uniform(vec1.y, vec2.y),
            uniform(vec1.z, vec2.z)
        ))
        return vec.copy()

    def get_bool(self):
        return random() <= self.get_input(self.probability)



from uplogic.nodes import ULParameterNode
import bge


class ULGetObject(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.OUT = self.add_output(self.get_obj)
        self.scene = bge.logic.getCurrentScene()

    def get_obj(self):
        return self.get_input(self.game_object)



from uplogic.nodes import ULConditionNode
from uplogic.input import mouse_moved


class ULMouseMoved(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        if self.pulse:
            return mouse_moved()
        return mouse_moved(tap=True)



from bge import logic
from uplogic.nodes import ULActionNode


class ULCursorBehavior(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.cursor_object = None
        self.world_z = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        cursor_object = self.get_input(self.cursor_object)
        if not self.get_condition():
            if cursor_object.visible:
                cursor_object.setVisible(False, True)
            return
        camera = logic.getCurrentScene().active_camera
        world_z = self.get_input(self.world_z)
        if not cursor_object.visible:
            cursor_object.setVisible(True, True)
        else:
            x = self.network.mouse.position[0]
            y = self.network.mouse.position[1]
            direction = camera.getScreenVect(x, y)
            origin = camera.worldPosition
            aim = direction * -world_z
            point = origin + aim
            cursor_object.worldOrientation = camera.worldOrientation
            cursor_object.worldPosition = point
        self._done = True



from uplogic.nodes import ULConditionNode
from bge.logic import getRealTime


class ULPulsify(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = None
        self.delay = None
        self.result = False
        self._old_time = getRealTime()
        self._time = 0.0
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self.result

    def evaluate(self):
        self.result = False
        now = getRealTime()
        if not self.get_input(self.condition):
            self._time = 0.0
            self._old_time = now
            return
        if self._time <= 0.0:
            self.result = True
            self._time = self.get_input(self.delay)
        self._time -= now - self._old_time
        self._old_time = now
        


from bge import logic


class Portal:
    
    def __init__(self) -> None:
        self.node = None
    
    @property
    def value(self):
        if self.node is not None:
            return self.node.value
        return None


class Portals:

    _portals = {}

    @classmethod
    def get(cls, key):
        portal = cls._portals.get(key, None)
        if portal is None:
            portal = Portal()
            cls._portals[key] = portal
        return portal



from mathutils import Vector
from uplogic.nodes import ULParameterNode


class ULObjectDataVertices(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.OUT = self.add_output(self.get_data)

    def get_data(self):
        obj = self.get_input(self.game_object)
        offset = obj.worldPosition
        return sorted(
            [Vector(v.co) + offset for v in (
                obj
                .blenderObject
                .data
                .vertices
            )]
        )



from bge import render
from bge import logic
from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.input import MouseLook
from math import degrees


class ULMouseLook(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.axis = None
        self.condition = None
        self.game_object_x = None
        self.game_object_y = None
        self.inverted = None
        self.sensitivity = None
        self.use_cap_z = None
        self.cap_z = None
        self.use_cap_y = None
        self.cap_y = None
        self.smooth = None
        self.initialized = False
        self.center_mouse = True
        self.use_local_head = False
        self._x = 0
        self._y = 0
        self.OUT = self.add_output(self.get_done)
        self.mouselook = None

    def reset(self):
        super().reset()
        if self.mouselook:
            self.mouselook.get_data()

    def evaluate(self):
        condition = self.get_condition()
        if condition:
            obj = self.get_input(self.game_object_x)
            head = self.get_input(self.game_object_y)
            if self.mouselook and self.mouselook.active:
                if (
                    self.mouselook.obj is not obj
                    or
                    head is not None and self.mouselook.head is not head
                ):
                    self.mouselook.disable()
                    self.mouselook = None
            if self.mouselook is None:
                use_cap_x = self.get_input(self.use_cap_z)
                use_cap_y = self.get_input(self.use_cap_y)
                cap_x = [degrees(x) for x in self.get_input(self.cap_z)] if use_cap_x else None
                cap_y = [degrees(y) for y in self.get_input(self.cap_y)] if use_cap_y else None
                self.mouselook = MouseLook(
                    obj=obj,
                    head=head,
                    sensitivity=self.get_input(self.sensitivity),
                    use_cap_x=use_cap_x,
                    cap_x=cap_x,
                    use_cap_y=use_cap_y,
                    cap_y=cap_y,
                    invert=self.get_input(self.inverted),
                    smoothing=self.get_input(self.smooth),
                    local=head is not obj,
                    center_mouse=self.center_mouse,
                    front=self.get_input(self.axis),
                    active=False
                )
            self._done = True
        if self.mouselook is not None and condition:
            self.mouselook.update()
            self._done = True



from uplogic.nodes import ULActionNode


class ULCopyProperty(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.from_object = None
        self.to_object = None
        self.property_name = None
        self.mode = 0
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        from_object = self.get_input(self.from_object)
        to_object = self.get_input(self.to_object)
        property_name = self.get_input(self.property_name)
        from_object = from_object.blenderObject if self.mode else from_object
        to_object = from_object.blenderObject if self.mode else to_object
        val = from_object.get(property_name)
        if val is not None:
            to_object[property_name] = val
            self._done = True



'''TODO: Documentation
'''

from bge import logic
import time
# from uplogic.physics import on_collision  # noqa


def on_pre_draw(callback):
    EventManager.update_on.append(callback)


def on_post_draw(callback):
    EventManager.update_on.append(callback)


class _DeferredCallback:


    def callback(self):
        ...

    def __init__(self, callback) -> None:
        self.callback = callback
        logic.getCurrentScene().post_draw.append(self._call)

    def _call(self):
        self.callback()
        if self._call in logic.getCurrentScene().post_draw:
            logic.getCurrentScene().post_draw.remove(self._call)


def later(callback):
    class _DeferredCallback():

        def __init__(self, callback) -> None:
            self.callback = callback
            logic.getCurrentScene().post_draw.append(self._call)

        def _call(self):
            self.callback()
            logic.getCurrentScene().post_draw.remove(self._call)
    _DeferredCallback(callback)


def get_event_manager():
    update = EventManager.update_on
    if update is None:
        update = logic.getCurrentScene().post_draw
    if EventManager.update not in update:
        update.append(EventManager.update)
    return EventManager


def set_update_loop(loop):
    EventManager.set_update_on(loop)


class EventManager:
    '''Manager for `Event` objects, not inteded for manual use.
    '''
    events = {}
    bindings = []
    scheduled = []
    done = []

    update_on = None

    @classmethod
    def set_update_on(cls, li):
        old = cls.update_on
        if li is None:
            li = logic.getCurrentScene().post_draw
        if old is None:
            old = logic.getCurrentScene().post_draw
        if cls.update in old:
            old.remove(cls.update)
        if cls.update not in li:
            li.append(cls.update)
        cls.update_on = li

    @classmethod
    def update(cls):
        cbs = cls.bindings.copy()
        cbs.extend(cls.scheduled)
        for cb in cbs:
            cb()

    @classmethod
    def log(cls):
        if cls.events:
            print('Events:')
            for evt in cls.events:
                print(f'\t{evt}:\t{cls.events[evt].content}')

    @classmethod
    def schedule(cls, cb):
        get_event_manager()
        cls.scheduled.append(cb)

    @classmethod
    def cancel(cls, cb):
        get_event_manager()
        if cb in cls.scheduled:
            cls.scheduled.remove(cb)

    @classmethod
    def bind(cls, cb):
        get_event_manager()
        cls.bindings.append(cb)

    @classmethod
    def unbind(cls, cb):
        get_event_manager()
        if cb in cls.bindings:
            cls.bindings.remove(cb)

    @classmethod
    def release(cls, cb):
        get_event_manager()
        cls.cancel(cb)

    @classmethod
    def register(cls, event):
        get_event_manager()
        cls.events[event.id] = event
        cls.schedule(event.remove)

    @classmethod
    def send(cls, id, content, messenger, target) -> None:
        get_event_manager()
        Event(id, content, messenger, target)

    @classmethod
    def receive(cls, id, target=None):
        get_event_manager()
        evt = cls.events.get(id, None)
        if evt:
            if evt.target is None:
                return evt
            elif evt.target is not None and evt.target is target:
                return evt
        return None
        # return evt if evt and target is evt.target else None

    @classmethod
    def consume(cls, id):
        get_event_manager()
        return cls.events.pop(id, None)

    @classmethod
    def clear_schedule(cls):
        cls.scheduled.clear()

    @classmethod
    def clear_bindings(cls):
        cls.bindings.clear()

class Event():
    '''Event generated by `uplogic.events.send()`.

    **Not intended for manual use.**

    :param id: Identifier of the event; can be any object.
    :param content: This can be used to store data in an event.
    :param messenger: Can be used to store an object.
    '''

    def __init__(self, id: int, content=None, messenger=None, target=None):
        self.id = id
        self.content = content
        self.messenger = messenger
        self.target = target
        EventManager.schedule(self.register)

    def register(self):
        EventManager.register(self)
        EventManager.cancel(self.register)

    def remove(self):
        EventManager.events.pop(self.id, None)
        EventManager.cancel(self.remove)


def clear_schedule():
    EventManager.clear_schedule()


def send(id: int, content=None, messenger=None, target=None) -> None:
    '''Send an event that can be reacted to.

    :param id: Identifier of the event; can be anything, not just `int`.
    :param content: This can be used to store data in an event.
    :param messenger: Can be used to store an object.
    :param target: Target object. If defined, only target object can receive.
    '''
    EventManager.send(id, content, messenger, target)


def receive(id, target=None) -> Event:
    '''Check if an event has occured.

    :param id: Identifier of the event; can be any object.
    :param target: Target object. If defined, only target object can receive.

    :returns: `Event` with `id`, `content` and `messenger` as attributes.
    '''
    return EventManager.receive(id, target)


def consume(id: int):
    '''Check if an event has occured. This will remove the event.

    :param id: Identifier of the event; can be any object.

    :returns: `Event` with `id`, `content` and `messenger` as attributes.
    '''
    return EventManager.consume(id)


def bind(id: int, callback, *args):
    '''Bind a callback to an event. This callback will be executed whenever
    the specified event is received.

    **Required signature**:
    ```
    def callback(evt):
        ...
    ```

    :param id: Identifier of the event; can be any object.
    :param callback: This callback will be called every time the event is
    triggered.
    '''
    class _BoundCallback():
        def __init__(self, id, cb, *args) -> None:
            self.id = id
            self.callback = cb
            self.args = args
            EventManager.bind(self._check_evt)

        def _check_evt(self):
            evt = receive(self.id)
            if evt:
                self.callback(evt, *self.args)

        def unbind(self):
            EventManager.unbind(self._check_evt)

        def release(self):
            EventManager.unbind(self._check_evt)

    return _BoundCallback(id, callback, *args)


def bind_once(id: int, callback, *args):
    '''Bind a callback to an event and automatically unbind when received.

    Required signature: `def cb(evt)`

    :param id: Identifier of the event; can be any object.
    :param callback: This callback will be called every time the event is
    triggered.
    '''
    class _BoundCallback():
        def __init__(self, id, cb, *args) -> None:
            self.id = id
            self.callback = cb
            self.args = args
            EventManager.bind(self._check_evt)

        def _check_evt(self):
            evt = receive(self.id)
            if evt:
                self.callback(evt, *self.args)
                self.unbind()

        def unbind(self):
            EventManager.unbind(self._check_evt)

        def release(self):
            EventManager.unbind(self._check_evt)

    return _BoundCallback(id, callback, *args)


class ULEvent(Event):
    # XXX: Remove legacy support
    pass


class ScheduledEvent():
    '''Event generated by `uplogic.events.schedule()`.

    :param delay: Delay with which to send the event in seconds.
    :param id: Identifier of the event; can be any object
    :param content: This can be used to store data in an event.
    :param messenger: Can be used to store an object.
    '''

    def __init__(self, delay, id, content, messenger):
        self.time = time.time()
        self.delay = self.time + delay
        self.id = id
        self.content = content
        self.messenger = messenger
        self._consumed = False
        EventManager.schedule(self._send_scheduled)

    @property
    def time_left(self):
        return self.delay - time.time()

    def _send_scheduled(self):
        if time.time() >= self.delay:
            self.consume()

    def cancel(self):
        EventManager.cancel(self._send_scheduled)

    def consume(self):
        if self._consumed:
            return
        self.delay = time.time()
        self._consumed = True
        EventManager.cancel(self._send_scheduled)
        Event(self.id, self.content, self.messenger)


class _ScheduledEvent(ScheduledEvent):
    # XXX: Remove legacy support
    pass


def schedule(id: str, delay=0.0, content=None, messenger=None) -> ScheduledEvent:
    '''Send an event that can be reacted to with a delay.

    :param id: Identifier of the event; can be any object. If `id` is callable, `content` can be used as argument.
    :param delay: Time delay in seconds.
    :param content: This can be used to store data in an event.
    :param messenger: Can be used to store an object.
    '''
    if callable(id):
        return ScheduledCallback(id, delay, content)
    return ScheduledEvent(delay, id, content, messenger)


class ScheduledCallback():
    '''Event generated by `uplogic.events.schedule_callback()`.

    **Not intended for manual use.**

    :param cb: Callback to be evaluated.
    :param delay: Delay with which to call the function in seconds.
    :param arg: If this is defined, callback will be called with this
    argument.
    '''

    def __init__(self, cb, delay=0.0, arg=None):
        self.time = time.time()
        self.delay = self.time + delay
        self.callback = cb
        self.arg = arg
        self._consumed = False
        EventManager.schedule(self._call_scheduled)

    @property
    def time_left(self):
        return self.delay - time.time()

    def _call_scheduled(self):
        if time.time() >= self.delay:
            self.consume()

    def consume(self):
        if self._consumed:
            return
        self.delay = time.time()
        self._consumed = True
        EventManager.cancel(self._call_scheduled)
        if self.arg is not None:
            self.callback(self.arg)
        else:
            self.callback()

    def cancel(self):
        EventManager.cancel(self._call_scheduled)


class _ScheduledCallback(ScheduledCallback):
    # XXX: Remove legacy support
    pass


def schedule_callback(cb, delay=0.0, arg=None) -> ScheduledCallback:
    '''Call a function with a delay. The function can have an argument when
    defined as a keyword.

    Callback cannot return anything.

    :param cb: Callback to be evaluated.
    :param delay: Delay with which to call the function in seconds.
    :param arg: If this is defined, callback will be called with this
    argument.
    '''
    return ScheduledCallback(cb, delay, arg)



from uplogic.nodes import ULActionNode


class ULRemoveParent(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.child_object = None
        self._parent = None
        self.OUT = self.add_output(self.get_done)
        self.PARENT = self.add_output(self.get_parent)

    def get_done(self):
        return self._done

    def get_parent(self):
        return self._parent

    def reset(self):
        super().reset()
        self._parent = None

    def evaluate(self):
        if not self.get_condition():
            return
        child_object = self.get_input(self.child_object)
        self._parent = child_object.parent
        child_object.removeParent()
        self._done = True



from uplogic.nodes import ULConditionNode
from uplogic.utils.constants import LOGIC_OPERATORS


class ULCompare(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.operator = 'GREATER'
        self.param_a = None
        self.param_b = None
        self.threshold = 0.0
        self.RESULT = self.add_output(self.get_result)

    def get_result(self):
        a = self.get_input(self.param_a)
        b = self.get_input(self.param_b)
        threshold = self.get_input(self.threshold)
        operator = self.get_input(self.operator)
        if threshold is None:
            threshold = 0
        if threshold > 0 and abs(a - b) < threshold:
            a = b
        if operator is None:
            return
        return LOGIC_OPERATORS[operator](a, b)



from mathutils import Vector
from uplogic.nodes import ULParameterNode


class ULVectorLength(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_v = None
        self.output_v = Vector()
        self.OUTV = self.add_output(self.get_out_v)

    def get_out_v(self):
        return self.get_input(self.input_v).length



from bge import render
from uplogic.nodes import ULActionNode


class ULSetResolution(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.x_res = None
        self.y_res = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        x_res = self.get_input(self.x_res)
        y_res = self.get_input(self.y_res)
        render.setWindowSize(x_res, y_res)
        self._done = True



from uplogic.nodes import ULActionNode
from uplogic.physics.vehicle import Vehicle
from uplogic.utils.constants import VEHICLE


class ULCreateVehicle(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.suspension = None
        self.stiffness = None
        self.damping = None
        self.friction = None
        self.wheel_size = None
        self.vehicle = None
        self.OUT = self.add_output(self.get_done)
        self.VEHICLE = self.add_output(self.get_vehicle)
        self.WHEELS = self.add_output(self.get_wheels)

    def get_done(self):
        return self._done

    def get_vehicle(self):
        return self.vehicle

    def get_wheels(self):
        return self.vehicle.wheels

    def evaluate(self):
        game_object = self.get_input(self.game_object)
        if not self.get_condition():
            if game_object.get(VEHICLE):
                self.vehicle = game_object[VEHICLE]
            return
        suspension = self.get_input(self.suspension)
        stiffness = self.get_input(self.stiffness)
        damping = self.get_input(self.damping)
        friction = self.get_input(self.friction)
        wheel_size = self.get_input(self.wheel_size)
        self.vehicle = Vehicle(
            game_object,
            suspension,
            stiffness,
            damping,
            friction,
            wheel_size
        )
        self._done = True



from uplogic.nodes import ULParameterNode
from bpy.types import NodeTree


class ULGetNodeSocket(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.tree_name = None
        self.node_name = None
        self.input_slot = None
        self.OUT = self.add_output(self._get_val)

    def _get_val(self):
        tree: NodeTree = self.get_input(self.tree_name)
        node_name = self.get_input(self.node_name)
        input_slot = self.get_input(self.input_slot)
        return (
            tree
            .nodes[node_name]
            .inputs[input_slot]
            .default_value
        )



from bge import render
from uplogic.nodes import ULActionNode


class ULSetFullscreen(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.use_fullscreen = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        use_fullscreen = self.get_input(self.use_fullscreen)
        render.setFullScreen(use_fullscreen)
        self._done = True



from mathutils import Vector
from uplogic.nodes import ULParameterNode


class VectorNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.vector = Vector()
        self.OUT = self.add_output(self.get_vec)

    def get_vec(self):
        return self.vector



from uplogic.nodes import ULParameterNode
from mathutils import Vector


class ULClamp(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.range = None
        self.min_value = None
        self.max_value = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        value = self.get_input(self.value)
        if self.min_value is not None:
            range_ft = Vector((self.get_input(self.min_value), self.get_input(self.max_value)))
        else:
            range_ft = self.get_input(self.range)

        if range_ft.x == range_ft.y:
            return value
        if value < range_ft.x:
            value = range_ft.x
        if value > range_ft.y:
            value = range_ft.y
        return value



from uplogic.nodes import ULParameterNode
from inspect import signature
from uplogic.utils.math import clamp


class MathNode(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.value_a = 0.0
        self.value_b = 0.0
        self.value_c = 0.0
        self.clamp = False
        self.operator = None
        self.RESULT = self.add_output(self.get_result)

    def get_result(self):
        a = self.get_input(self.value_a)
        b = self.get_input(self.value_b)
        c = self.get_input(self.value_c)
        op = self.operator
        args = len(signature(op).parameters)
        if args == 1:
            result = self.operator(a)
        elif args == 2:
            result = self.operator(a, b)
        else:
            result = self.operator(a, b, c)
        return clamp(result) if self.clamp else result



from uplogic.nodes import ULActionNode
from uplogic.data import GlobalDB


class ULStopAllSounds(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        aud_sys = GlobalDB.retrieve('uplogic.audio').get('default')
        aud_sys.device.stopAll()
        self._done = True



from uplogic.nodes import ULParameterNode
from uplogic.utils import get_project_folder
import bpy


class GetMasterFolderNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.name = ''
        self.PATH = self.add_output(self._get_path)

    def _get_path(self):
        name = self.get_input(self.name)

        return get_project_folder(name)



from .layout import RelativeLayout
from .layout import ScrollBoxLayout
from .slider import FrameSlider
from uplogic.input.mouse import mouse_wheel
from uplogic.utils.math import map_range
from uplogic.utils.math import clamp
from bge import logic
from bge import render


class ScrollView(RelativeLayout):

    def __init__(
        self,
        pos: list = [0, 0],
        size: list = [100, 100],
        bg_color: list = [0, 0, 0, 0],
        relative: dict = {},
        border_width: int = 1,
        border_color: list = [0, 0, 0, 0],
        inverted: bool = False,
        spacing: int = 0,
        halign: str = 'left',
        valign: str = 'bottom',
        scroll_speed=30,
        use_scroll_bar=True,
        scroll_bar_width=5,
        scroll_bar_color=(0, 0, 0, .8),
        scroll_bar_knob_color=(.8, .5, .1, 1),
        angle=0,
        show=True
    ):
        self.scroll_box = ScrollBoxLayout(orientation='vertical', relative={'size': True}, size=(1, 1), spacing=spacing, inverted=inverted)
        super().__init__(pos, size, bg_color, relative, border_width, border_color, halign, valign, angle, show)
        self.scroll_speed = scroll_speed
        self.use_scroll_bar = use_scroll_bar
        self.scroll_bar = FrameSlider(orientation='vertical', pos=(1, 0), relative={'pos': True}, halign='right', size=(0, 0), bg_color=(0, 0, 0, 0), border_color=(0, 0, 0, 0), bar_hover_color=scroll_bar_color)
        self.scroll_bar.square_knob = False
        self.scroll_bar_width = scroll_bar_width
        self.scroll_bar.on_value = self._on_scroll_bar
        self.add_widget(self.scroll_box)
        self.add_widget(self.scroll_bar)
        self.scroll_box._count_children = self._count_children
        self.remove_widget = self.scroll_box.remove_widget
        self.add_widget = self.scroll_box.add_widget
        self.scroll = self.scroll_box.scroll
        self.scroll_bar.value = 0

    @property
    def spacing(self):
        return self.scroll_box.spacing

    @spacing.setter
    def spacing(self, val):
        self.scroll_box.spacing = val

    @property
    def seek_speed(self):
        return self.scroll_box.seek_speed

    @seek_speed.setter
    def seek_speed(self, val):
        self.scroll_box.seek_speed = val

    def _on_scroll_bar(self, widget):
        self.scroll_box.scroll_position = self.scroll_bar.value

    def _count_children(self):
        ScrollBoxLayout._count_children(self.scroll_box)
        self.scroll_bar.knob_size = [1, clamp(map_range(self.scroll_box._height_diff, 0, self.height_pixel * 3, 1, 0), 0.05, 1)]
        self.scroll_bar.show = self.scroll_box._height_diff >= 0 and self.use_scroll_bar
        self.scroll_bar.size = [self.scroll_bar_width, self.height_pixel]
        # self.scroll_bar.set_value(self.scroll_box.scroll_position_actual)

    # def evaluate(self):
    #     super().evaluate()

    def evaluate(self):
        mpos = logic.mouse.position
        if self.check_inside(mpos[0] * render.getWindowWidth(), mpos[1] * render.getWindowHeight()):
            self.scroll(mouse_wheel())
        if not self.scroll_bar.bar._down:
            self.scroll_bar.set_value(self.scroll_box.scroll_position_actual)

    def scroll(self, difference):
        if difference != 0:
            self.scroll_box.scroll(difference)
        


from uplogic.nodes import ULConditionNode
from uplogic.utils.math import clamp
from uplogic.utils.math import interpolate
from mathutils import Vector


class TweenValueNode(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.forward = False
        self.back = False
        self.from_float = 0.0
        self.to_float = 1.0
        self.from_vec = Vector((0, 0, 0))
        self.to_vec = Vector((1, 1, 1))
        self.duration = 1.0

        self._time = 0.0

        self.mapping = None
        self.active = False

        self.on_demand = True
        self.instant_reset = False
        self.value_type = 0  # 0 -> float, 1 -> Vector

        self._result = 0.0
        self._factor = 0.
        self._eval = 0.

        self._direction = 1

        self.DONE = self.add_output(self.get_done)
        self.REACHED = self.add_output(self.get_reached)
        self.RESULT_FLOAT = self.add_output(self.get_float)
        self.RESULT_VEC = self.add_output(self.get_vec)
        self.FAC = self.add_output(self.get_fac)

    def get_done(self):
        return self._done

    def get_reached(self):
        return self._factor == self._direction

    def get_fac(self):
        return self._factor

    def get_vec(self):
        self.active = True
        if self.on_demand:
            self.forward = True
            self.back = False
        self._result = self.get_input(self.from_vec).lerp(self.get_input(self.to_vec), self._eval)
        return self._result

    def get_float(self):
        self.active = True
        if self.on_demand:
            self.forward = True
            self.back = False
        self._result = interpolate(self.get_input(self.from_float), self.get_input(self.to_float), self._eval)
        return self._result

    def reset(self):
        self.active = False
        return super().reset()

    def evaluate(self):
        forward = self.get_condition(self.forward)
        
        back = self.get_condition(self.back)
        if not forward and not back:
            return
        self.active = True
        if not self.active and not self.on_demand:
            return
        duration = self.get_input(self.duration)
        mapping = self.mapping.curve
        if forward:
            self._direction = 1
            if duration > 0:
                self._time = clamp(self._time + self.network.time_per_frame, 0, duration)
                self._factor = self._time / duration
            else:
                self._factor = 1
            self._eval = clamp(mapping.evaluate(
                mapping.curves[0],
                self._factor
            ))
        elif back:
            self._direction = 0
            if self.on_demand and self.instant_reset:
                duration = 0
                self._time = 0.0
            if duration > 0:
                self._time = clamp(self._time - self.network.time_per_frame, 0, duration)
                self._factor = self._time / duration
            else:
                self._factor = 0
            self._eval = clamp(mapping.evaluate(
                mapping.curves[0],
                self._factor
            ))
        self.active = False
        if self.on_demand:
            self.forward = False
            self.back = True
        self._done = True



from uplogic.nodes import ULParameterNode
from mathutils import Vector


class ULRangedThreshold(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.threshold = None
        self.min_value = None
        self.max_value = None
        self.operator = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        v = self.get_input(self.value)
        if self.min_value is not None:
            t = Vector((self.get_input(self.min_value), self.get_input(self.max_value)))
        else:
            t = self.get_input(self.threshold)
        value = self.calc_threshold(self.operator, v, t)
        return value

    def calc_threshold(self, op, v, t):
        if op == 'OUTSIDE':
            return v if (v < t.x or v > t.y) else 0
        if op == 'INSIDE':
            return v if (t.x < v < t.y) else 0



from uplogic.nodes import ULActionNode


class FModSetEventAttributeNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = False
        self.event = None
        self.attribute = 'position'
        self.value = 0.0
        self.EVT = self.add_output(self.get_evt)
        self.VALUE = self.add_output(self.get_value)

    def get_evt(self):
        return self.get_input(self.event)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        setattr(
            self.get_input(self.event),
            self.attribute,
            self.get_input(self.value)
        )



# from bge import render, logic
from bge.render import drawLine
from mathutils import Vector
from bge.types import KX_GameObject
from bpy.types import Mesh, MeshEdge
from uplogic import console


def draw_line(origin: Vector, target: Vector, color: list = [1, 1, 1, 1]):
    console.warning('Moved from "uplogic.utils.visuals" to "uplogic.utils.visualize"!')
    drawLine(
        origin,
        target,
        color
    )


def draw_path(points: list, color: list = [1, 1, 1, 1]):
    console.warning('Moved from "uplogic.utils.visuals" to "uplogic.utils.visualize"!')
    for i, p in enumerate(points):
        if i < len(points) - 1:
            drawLine(p, points[i+1], color)


def draw_cube(origin: Vector, width: float = 1, color: list = [1, 1, 1, 1], centered: bool = False):
    console.warning('Moved from "uplogic.utils.visuals" to "uplogic.utils.visualize"!')
    draw_box(origin, width, width, width, color, centered)


def draw_box(origin: Vector, width: float, length: float, height: float, color: list = [1, 1, 1, 1], centered: bool = False):
    console.warning('Moved from "uplogic.utils.visuals" to "uplogic.utils.visualize"!')
    if centered:
        origin = origin.copy() - Vector((width * .5, length * .5, height * .5))

    c1: Vector = origin.copy()
    c2: Vector = origin.copy()
    c3: Vector = origin.copy()
    c4: Vector = origin.copy()
    c5: Vector = origin.copy()
    c6: Vector = origin.copy()
    c7: Vector = origin.copy()

    c1[0] += length

    c2[0] += length
    c2[1] += width

    c3[1] += width

    c4[2] += height

    c5[0] += length
    c5[2] += height

    c6[0] += length
    c6[1] += width
    c6[2] += height

    c7[1] += width
    c7[2] += height

    drawLine(origin, c1, color)
    drawLine(c1, c2, color)
    drawLine(c2, c3, color)
    drawLine(c3, origin, color)

    drawLine(origin, c4, color)
    drawLine(c1, c5, color)
    drawLine(c2, c6, color)
    drawLine(c3, c7, color)

    drawLine(c4, c5, color)
    drawLine(c5, c6, color)
    drawLine(c6, c7, color)
    drawLine(c7, c4, color)


def draw_mesh(game_object: KX_GameObject, color: tuple = (1, 1, 1, 1)):
    console.warning('Moved from "uplogic.utils.visuals" to "uplogic.utils.visualize"!')
    mesh: Mesh = game_object.blenderObject.data
    for edge in mesh.edges:
        v1 = mesh.vertices[edge.vertices[0]]
        v2 = mesh.vertices[edge.vertices[1]]
        drawLine(
            game_object.worldTransform @ Vector(v1.co),
            game_object.worldTransform @ Vector(v2.co),
            color
        )



from uplogic.nodes import ULParameterNode
import sys
import random


class ULRandomFloat(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.max_value = None
        self.min_value = None
        self.OUT_A = self.add_output(self._get_output)

    def _get_output(self):
        min_value = self.get_input(self.min_value)
        max_value = self.get_input(self.max_value)
        if min_value > max_value:
            min_value, max_value = max_value, min_value
        if min_value == max_value:
            min_value = sys.float_info.min
            max_value = sys.float_info.max

        delta = max_value - min_value
        return min_value + (delta * random.random())



from uplogic.nodes import ULParameterNode
from mathutils import Vector, Matrix
import bpy
from bge import logic


class ULRebuildData(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.condition = True
        self.data = None
        self.read_as = 'builtin'
        self.ret = None
        self.dattypes = {
            'Vec2': Vector,
            'Vec3': Vector,
            'Vec4': Vector,
            'Mat3': Matrix,
            'Mat4': Matrix
        }
        self.OUT = self.add_output(self.get_data)

    def get_data(self):
        return self.ret

    def fetch(self):
        if not self.get_input(self.condition):
            return
        dat = self.get_input(self.data)
        if dat is None:
            return
        if self.read_as == 'builtin':
            self.ret = dat
            return
        if self.read_as == 'GameObj':
            bobj = bpy.data.objects.get(dat['name'])
            if bobj is None:
                self.ret = None
                return
            obj = logic.getCurrentScene().getGameObjectFromObject(bobj)
            for attr in dat:
                if attr == 'properties':
                    for prop in dat[attr]:
                        obj[prop] = dat[attr][prop]
                elif attr not in ['name', 'data_id']:
                    setattr(obj, attr, dat[attr])
            self.ret = obj
            return
        self.ret = self.dattypes[self.read_as](dat)



from uplogic.nodes import ULConditionNode
from uplogic.nodes import LoopList


class LoopNode(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = False
        self.get_input = self._get_input
        self.items = []
        self.LOOP = self.add_output(self.get_loop)
        self.LIST = self.add_output(self.get_list)

    @property
    def loop_mode(self):
        return self._loop_mode

    @loop_mode.setter
    def loop_mode(self, val):
        self._loop_mode = False

    def get_loop(self):
        cond = self.get_input(self.condition)
        return LoopList([cond for x in self.get_input(self.items)])

    def get_list(self):
        return LoopList([self.get_input(item) for item in self.get_input(self.items)])



from uplogic.nodes import ULParameterNode


class CurveInterpolationNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.mapping = None
        self.OUT = self.add_output(self.get_result)

    def get_result(self):
        val = self.get_input(self.value)
        mapping = self.mapping.curve

        return mapping.evaluate(mapping.curves[0], val)



from mathutils import Vector
from uplogic.nodes import ULActionNode
import random


class ULRandomVect(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.xyz = None
        self.OUT_A = self.add_output(self._get_output)

    def _get_output(self):
        xyz = self.get_input(self.xyz)
        vmin, vmax = -999999999, 999999999
        delta = vmax - vmin
        x = vmin + (delta * random.random()) if xyz['x'] else 0
        y = vmin + (delta * random.random()) if xyz['y'] else 0
        z = vmin + (delta * random.random()) if xyz['z'] else 0
        return Vector((x, y, z))



from .widget import Widget
import gpu
import bpy
from math import ceil
from uplogic.utils import clamp
from .widget import rotate2d
from gpu_extras.batch import batch_for_shader
from mathutils import Vector
from os.path import isfile
from uplogic.utils.handlers import ImageHandler



class _UV_Point(list):

    @property
    def lower(self):
        return self[0]

    @lower.setter
    def lower(self, val):
        self[0] = val

    @property
    def upper(self):
        return self[1]

    @upper.setter
    def upper(self, val):
        self[1] = val


class _UV(list):

    @property
    def owner(self) -> Widget:
        return self[2]

    @property
    def x(self) -> _UV_Point:
        return self[0]

    @x.setter
    def x(self, val):
        self[0] = _UV_Point(val)
        self.owner._build_shader()

    @property
    def x_min(self) -> _UV_Point:
        return self.x.lower

    @x_min.setter
    def x_min(self, val):
        self.x.lower = val
        self.owner._build_shader()

    @property
    def x_max(self) -> _UV_Point:
        return self.x.upper

    @x_max.setter
    def x_max(self, val):
        self.x.upper = val
        self.owner._build_shader()

    @property
    def y(self) -> _UV_Point:
        return self[1]

    @y.setter
    def y(self, val):
        self[1] = _UV_Point(val)
        self.owner._build_shader()

    @property
    def y_min(self) -> _UV_Point:
        return self.y.lower

    @y_min.setter
    def y_min(self, val):
        self.y.lower = val
        self.owner._build_shader()

    @property
    def y_max(self) -> _UV_Point:
        return self.y.upper

    @y_max.setter
    def y_max(self, val):
        self.y.upper = val
        self.owner._build_shader()


class Image(Widget):

    vertex_in = [
        ('VEC2', 'texCoord'),
        ('VEC2', 'position')
    ]

    interfaces = [
        ('VEC2', "pos"),
        ('VEC2', 'uv')
    ]

    constants = [
        ('FLOAT', 'alpha')
    ]

    samplers = [
        ('FLOAT_2D', 'image')
    ]

    vertex_shader = """

    void main()
    {
        uv = texCoord;
        pos = position;
        gl_Position = ModelViewProjectionMatrix * vec4(pos.xy, 0.0, 1.0);
    }
    """

    fragment_shader = """

    void main()
    {
        vec4 color = mix(vec4(0.0), texture(image, uv), alpha);
        FragColor = pow(color, vec4(0.5));
    }
    """

    def __init__(
        self,
        pos=[0, 0],
        size=(100, 100),
        relative={},
        texture=None,
        halign='left',
        valign='bottom',
        use_aspect_ratio: bool = True,
        angle=0,
        show=True
    ):
        self._texture = None
        self.use_aspect_ratio = use_aspect_ratio
        self._uv: _UV[_UV_Point] = _UV((_UV_Point((0.001, .999)), _UV_Point((0.001, .999)), self))
        self._opacity = 1
        self._load_image(texture)
        super().__init__(pos, size, relative=relative, halign=halign, valign=valign, angle=angle, show=show)

    @property
    def uv(self):
        return self._uv

    @uv.setter
    def uv(self, val):
        self._uv = val
        self._build_shader()

    @property
    def image(self):
        return self.image_handler.image

    def _load_image(self, texture):
        self.image_handler = ImageHandler(texture)
        self.start()

    @property
    def size_pixel(self):
        if self._vertices is None:
            return [0, 0]
        bottom_left = self._vertices[1]
        top_right = self._vertices[2]
        return [top_right[0] - bottom_left[0], top_right[1] - bottom_left[1]]

    @property
    def filepath(self):
        return self.image_handler.filepath

    @property
    def texture(self):
        return self.image_handler._texture

    @property
    def aspect_ratio(self):
        if self.image is None:
            return 1
        return self.image.size[1] / self.image.size[0]

    @property
    def _draw_size(self):
        size = self.size.copy()
        use_aspect_ratio = self.use_aspect_ratio
        if self.parent is None:
            return size
        if use_aspect_ratio and self.image:
            size[1] = size[0] * self.aspect_ratio
        if self.relative.get('size'):
            pdsize = self.parent._draw_size
            size = [
                size[0] * pdsize[0],
                size[1] * (pdsize[0] if use_aspect_ratio else pdsize[1])
            ]
        return size

    @texture.setter
    def texture(self, val):
        self.image_handler.texture = val
        self.size = self.size

    def free(self):
        self.image_handler.free()

    @property
    def frame(self):
        return self.image_handler.frame

    @frame.setter
    def frame(self, val):
        self.image_handler.frame = val

    @property
    def max_frame(self):
        return self.image_handler.max_frame

    @property
    def pivot(self):
        """Rotation point for this widget."""
        if self.parent is None:
            return (0, 0)
        v = self._vertices
        x0 = Vector(v[1])
        x1 = Vector(v[0])
        y1 = Vector(v[2])
        y0 = Vector(v[3])
        return Vector(self._get_pivot(x0, x1, y0, y1))

    def _build_shader(self):
        pos = self._draw_pos
        size = self._draw_size
        x0 = Vector((pos[0], pos[1]))
        x1 = Vector((pos[0] + size[0], pos[1]))
        y1 = Vector((pos[0] + size[0], pos[1] + size[1]))
        y0 = Vector((pos[0], pos[1] + size[1]))
        pivot = self._get_pivot(x0, x1, y0, y1)

        if self._draw_angle and self._vertices is not None:
            x0 = rotate2d(x0, pivot, self._draw_angle)
            x1 = rotate2d(x1, pivot, self._draw_angle)
            y0 = rotate2d(y0, pivot, self._draw_angle)
            y1 = rotate2d(y1, pivot, self._draw_angle)
        vertices = self._vertices = (
            x1, x0, y1, y0
        )

        self._shader = self._get_shader()


        uvs = self.uv
        self._batch = batch_for_shader(
            self._shader, 'TRI_STRIP',
            {
                "position": vertices,
                "texCoord": (
                    (uvs.x_max, uvs.y_min),
                    (uvs.x_min, uvs.y_min),
                    (uvs.x_max, uvs.y_max),
                    (uvs.x_min, uvs.y_max)
                ),
            },
        )
        if bpy.app.version[0] >= 4:
            self._shader.uniform_float("alpha", self.opacity)

    def draw(self):
        gpu.state.blend_set("ALPHA")
        self._setup_draw()
        if self.texture is None:
            super().draw()
            return
        self._shader.uniform_sampler("image", self.texture)
        self._shader.bind()
        self._batch.draw(self._shader)
        super().draw()


class Sprite(Image):

    def __init__(
        self,
        pos=[0, 0],
        size=(100, 100),
        relative={},
        texture=None,
        idx=0,
        rows=1,
        cols=1,
        halign='left',
        valign='bottom',
        use_aspect_ratio=True,
        show=True
    ):
        # self._idx = idx
        self.rows = rows
        self.cols = cols
        super().__init__(pos, size, relative, texture, halign=halign, valign=valign, use_aspect_ratio=use_aspect_ratio, show=show)
        self.idx = idx

    @property
    def idx(self):
        return self._idx

    @idx.setter
    def idx(self, val):
        self._idx = val
        self._get_uv()

    def _get_uv(self):
        ...
        idx = self.idx
        col = idx % self.cols
        col_end = col + 1
        row = ceil((idx + 1) / self.cols) - 1
        row_end = row + 1
        self.uv.y_max = 1 - row * self._row_height
        self.uv.y_min = 1 - row_end * self._row_height
        self.uv.x_max = col_end * self._col_width
        self.uv.x_min = col * self._col_width

    @property
    def aspect_ratio(self):
        if self.image is None:
            return 1
        return (self.image.size[1] / self.rows) / (self.image.size[0] / self.cols)

    @property
    def rows(self):
        return self._rows

    @rows.setter
    def rows(self, val):
        if val < 1:
            return
        self._rows = int(val)
        self._row_height = 1 / int(val)

    @property
    def cols(self):
        return self._cols

    @cols.setter
    def cols(self, val):
        if val < 1:
            return
        self._cols = int(val)
        self._col_width = 1 / int(val)

class Video(Image):

    def __init__(self, pos=[0, 0], size=(100, 100), relative={}, texture=None, halign='left', valign='bottom', use_aspect_ratio = True, fps=60, min_frame=0, max_frame=None, load_audio=True, play_mode='play', angle=0, show=True):
        self._load_audio = load_audio
        # self.play_mode = play_mode
        super().__init__(pos, size, relative, texture, halign, valign, use_aspect_ratio, angle, show)
        self.image_handler.fps = fps
        self.image_handler._min_frame = min_frame
        if max_frame is not None:
            self.image_handler._max_frame = max_frame

    def _load_image(self, texture):
        self.image_handler = ImageHandler(texture, load_audio=self._load_audio)

    @property
    def play_mode(self):
        return self.image_handler.play_mode

    @play_mode.setter
    def play_mode(self, val):
        self.image_handler.play_mode = val

    @property
    def texture(self):
        return self.image_handler.texture

    @texture.setter
    def texture(self, val):
        self.image_handler.texture = val

    @property
    def fps(self):
        return self.image_handler.fps

    @fps.setter
    def fps(self, val):
        self.image_handler.fps = val

    @property
    def playback_position(self):
        return self.image_handler.playback_position

    @playback_position.setter
    def playback_position(self, val):
        self.image_handler.playback_position = val

    @property
    def is_playing(self):
        return self.image_handler.is_playing

    @is_playing.setter
    def is_playing(self, val):
        self.image_handler.is_playing = val

    def play(self):
        self.image_handler.play()

    def seek(self, position):
        self.image_handler.seek(position)

    def stop(self):
        self.image_handler.stop()

    def remove(self):
        self.free()
        return super().remove()



from bge.render import drawLine
from uplogic.nodes import ULActionNode


class ULDrawLine(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.color = None
        self.from_point = None
        self.to_point = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        from_point = self.get_input(self.from_point)
        to_point = self.get_input(self.to_point)
        color = self.get_input(self.color)
        drawLine(
            from_point,
            to_point,
            [
                color.x,
                color.y,
                color.z,
                1
            ]
        )
        self._done = True



from uplogic.nodes import ULParameterNode
from uplogic.utils.math import map_range, map_range_vector


class ULMapRange(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.from_min = None
        self.from_max = None
        self.to_min = None
        self.to_max = None
        self.clamp = False
        self.mode = 0
        self.operations = [map_range, map_range_vector]
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self.operations[self.mode](
            self.get_input(self.value),
            self.get_input(self.from_min),
            self.get_input(self.from_max),
            self.get_input(self.to_min),
            self.get_input(self.to_max),
            self.clamp
        )



from uplogic.nodes import ULConditionNode


class ULOnce(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.input_condition = None
        self.condition = False
        self.repeat = None
        self.reset_time = None
        self._consumed = False
        self._result = False
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self._result

    def evaluate(self):
        self._result = False
        condition = self.get_input(self.input_condition)
        repeat = self.get_input(self.repeat)
        if condition and not self.condition:
            self._result = True
            self.condition = condition
        if repeat:
            self.condition = condition



import socket
import pickle
import threading
import bge
from uplogic.utils.constants import DISCONNECT_MSG
from uplogic import console
from uplogic.serialize import *

class Client:

    def __init__(self, server, port=8303, connect=False):
        self.scene = bge.logic.getCurrentScene()
        self.server = server
        self.socket = None
        self.port = port
        self.connected = False
        self.disconnect_on_scene_end = True
        if connect:
            self.connect()

    def connect(self):
        if self.connected:
            console.debug('Client Already Connected! Aborting.')
            return
        console.debug(f'Connecting to {self.server}...')
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            if self.disconnect_on_scene_end:
                if self.disconnect not in self.scene.onRemove:
                    self.scene.onRemove.append(self.disconnect)
            self.socket.connect((self.server, self.port))
            self.connected = True
            thread = threading.Thread(target=self.update)
            thread.start()
            console.success('[SUCCESS]')
        except socket.error as e:
            console.error('[ERROR]')
            console.error(e)
            return

    def disconnect(self, flag=True):
        if not self.connected or self.socket is None:
            console.debug('Client Not Connected!')
            return
        console.debug(f'Disconnecting from {self.server}...')
        if flag:
            self.socket.send(pickle.dumps(DISCONNECT_MSG))
            self.socket.shutdown(socket.SHUT_WR)
            self.socket.close()
        self.socket = None
        self.connected = False
        console.success('[SUCCESS]')

    def send(self, msg, subject=''):
        if self.connected and self.socket is not None:
            try:
                if subject:
                    msg = {
                        'subject': subject,
                        'content': msg
                    }
                self.socket.send(pickle.dumps(msg))
            except pickle.PicklingError:
                console.error(f'Cannot serialize {msg}!')
            except TypeError:
                console.error(f'Cannot serialize {msg}!')
            except socket.error:
                console.error('Server unreachable')
                self.disconnect()
            except Exception as e:
                console.error(f'Exception: {e}')
                self.disconnect()

    def on_receive(self, msg):
        pass

    def update(self):
        while self.connected and self.socket:
            try:
                bmsg = self.socket.recv(2048)
                msg = pickle.loads(bmsg)
                if not bmsg or msg == DISCONNECT_MSG:
                    self.connected = False
                else:
                    self.on_receive(msg)
            except Exception as e:
                self.connected = False
        self.disconnect(False)
        console.debug('Closing')



from .globaldb import GlobalDB
from .globaldb import retrieve  # noqa
from .globaldb import store  # noqa
from .file import read_file  # noqa
from .file import write_file  # noqa
from .file import load_file  # noqa
from .serializers import StringSerializer
from .serializers import FloatSerializer
from .serializers import IntegerSerializer
from .serializers import ListSerializer
from .serializers import VectorSerializer
from mathutils import Vector
from bge.types import KX_PythonComponent
from bge.types import KX_GameObject


GlobalDB.serializers[str(type(""))] = StringSerializer()
GlobalDB.serializers[str(type(1.0))] = FloatSerializer()
GlobalDB.serializers[str(type(10))] = IntegerSerializer()
GlobalDB.serializers[str(type([]))] = ListSerializer()
GlobalDB.serializers[str(type((0, 0, 0)))] = ListSerializer()
GlobalDB.serializers[str(type(Vector()))] = (
    VectorSerializer()
)


class GameProperty():

    def __init__(self, inst, name, default=None):
        
        def getPropComponent(self, attr_name=name):
                return self.object.get(attr_name)

        def setPropComponent(self, value, attr_name=name):
            self.object[attr_name] = value

        def getPropObject(self, attr_name=name):
            return self.get(attr_name)

        def setPropObject(self, value, attr_name=name):
            self[attr_name] = value

        if issubclass(inst.__class__, KX_PythonComponent):
            prop = property(getPropComponent, setPropComponent)
            if default:
                inst.object[name] = default
        elif issubclass(inst.__class__, KX_GameObject):
            prop = property(getPropObject, setPropObject)
            if default:
                inst[name] = default
        else:
            return
        setattr(inst.__class__, name, prop)


import bpy
from bge import logic
from uplogic import console


def init_glob_cats():
    # if not hasattr(bpy.types.Scene, 'nl_globals_initialized'):
    scene = logic.getCurrentScene()
    cats = getattr(
        bpy.data.scenes[scene.name],
        'nl_global_categories',
        None
    )
    if not cats:
        console.debug('No global categories found in', bpy.data.scenes[scene.name], bpy.data.scenes)
        return

    msg = ''

    dat = {
        '0': 'float_val',
        '1': 'string_val',
        '2': 'int_val',
        '3': 'bool_val',
        '17': 'filepath_val',
        '4': 'vec_val',
        '5': 'color_val',
        '6': 'color_alpha_val',
        '7': 'obj_val',
        '8': 'collection_val',
        '9': 'material_val',
        '10': 'mesh_val',
        '11': 'node_tree_val',
        '12': 'action_val',
        '13': 'text_val',
        '14': 'sound_val',
        '15': 'image_val',
        '16': 'font_val'
    }

    for c in cats:
        db = GlobalDB.retrieve(c.name)
        msg += f' {c.name},'
        for v in c.content:
            val = getattr(v, dat.get(v.value_type, '0'), 0)
            if isinstance(val, bpy.types.Object):
                val = logic.getCurrentScene().getGameObjectFromObject(val)
            elif int(v.value_type) in [4, 5, 6]:
                val = Vector(val)
            db.put(v.name, val, v.persistent)

    if msg:
        console.success(f'Globals Initialized:{msg[:-1]}')
    bpy.types.Scene.nl_globals_initialized = True


# if init_glob_cats not in bpy.app.handlers.load_post:
#     bpy.app.handlers.load_post.append(init_glob_cats)


from uplogic.nodes import ULConditionNode


class ULNot(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        condition = self.get_input(self.condition)
        return not condition



from bge import render
from uplogic.nodes import ULActionNode
from uplogic.ui import Cursor, remove_custom_cursor
from bpy.types import Image


class ULSetCustomCursor(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.texture = None
        self.size: list = None
        self._done: bool = None
        self._cursor = None
        self.OUT = self.add_output(self.get_done)
        self.CURSOR = self.add_output(self.get_cursor)

    def get_done(self):
        return self._done

    def get_cursor(self):
        return self._cursor

    def evaluate(self):
        if not self.get_condition():
            return
        remove_custom_cursor()
        texture: Image = self.get_input(self.texture)
        size = self.get_input(self.size)
        self._cursor = Cursor(texture=texture.name, size=size)
        self._done = True



from uplogic.nodes import ULActionNode
from uplogic.network import Client
from time import time
from uplogic.events import send, receive


class ULLocalClient(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.connect_cond = None
        self.ip_address = None
        self.port = None
        self.disconnect_cond = None

        self.on_init = False

        self._connect = False
        self._disconnect = False
        self._was_connected = False
        self._last_time = time()
        self._received = False
        self._message = None

        self.CONNECT = self.add_output(self.get_connect)
        self.CONNECTED = self.add_output(self.get_connected)
        self.DISCONNECT = self.add_output(self.get_disconnect)
        self.CLIENT = self.add_output(self.get_client)
        self.RECEIVED = self.add_output(self.get_received)
        self.MSG = self.add_output(self.get_message)

    def get_connect(self):
        return self._connect

    def get_connected(self):
        return self._client.connected

    def get_disconnect(self):
        return self._disconnect

    def get_client(self):
        return self._client

    def get_received(self):
        return self._received

    def get_message(self):
        return self._received.content if self._received else None

    def on_receive(self, msg):
        send(self._client, msg)

    def setup(self, tree):
        self._client = Client(
            self.get_input(self.ip_address),
            self.get_input(self.port),
            self.on_init
        )
        self._client.on_receive = self.on_receive

    def evaluate(self):
        self._connect = False
        self._disconnect = False
        self._received = receive(self._client)
        if self._was_connected != self._client.connected:
            if self._client.connected:
                self._connect = True
            else:
                self._disconnect = True
            self._was_connected = self._client.connected
        if self.get_input(self.connect_cond):
            self._client.connect()
        if self.get_input(self.disconnect_cond):
            self._client.disconnect()
        if not self._client.connected:
            self._client.server = self.get_input(self.ip_address)
            self._client.port = self.get_input(self.port)



from typing import Callable
from bge import logic
from bge.types import KX_GameObject as GameObject
from ..console import error


class Collision():
    """Callback handler for game object collisions.

    :param obj: Object whose collision detection will be monitored.
    :param callback: Callback to be called when collision occurs. Must have arguments `(obj, point, normal)`.
    :param prop: Only look for objects that have this property.
    :param material: Only look for objects that have this material applied.
    :param tap: Only validate the first frame of the collision.
    """

    _deprecated = False

    def __init__(
        self,
        game_object: GameObject,
        callback: Callable,
        prop: str = '',
        mat: str = '',
        tap: bool = False,
        post_call: bool = False
    ):
        if self._deprecated:
            from uplogic.console import warning
            warning('Warning: ULCollision class will be renamed to "Collision" in future releases!')
        self.point = None
        self.normal = None
        self.target = None
        self.consumed = False
        self.active = False
        self._old_target = None
        self._active = False
        self._objects = []
        self._old_objs = []
        self.callback: Callable = callback
        self.prop: str = prop
        self.mat: str = mat
        self.tap: bool = tap
        self.post_call = post_call
        self.game_object: GameObject = game_object
        self._done_objs = []
        self.register()

    def collision(self, obj, point, normal):
        if obj in self._objects:
            return
        material = self.mat
        prop = self.prop
        bo = obj.blenderObject
        if material:
            if material not in [
                slot.material.name for
                slot in
                bo.material_slots
            ]:
                return
        if prop:
            if prop not in obj.getPropertyNames():
                return

        self._objects.append(obj)
        self._active = True
        if obj not in self._old_objs:
            self.consumed = False
            self.target = obj
        self.active = not self.consumed if self.tap else True
        if self.active and obj not in self._done_objs:
            if (
                self.game_object.collisionGroup & obj.collisionMask and
                self.game_object.collisionMask & obj.collisionGroup
            ):
                self.callback(obj, point, normal)
                self.point = point
                self.normal = normal
        self._done_objs.append(obj)

    def reset(self):
        if self.post_call:
            for obj in self._old_objs:
                if obj not in self._done_objs:
                    self.callback(None, None, None)

        self.consumed = self._active
        self._active = False

        self._old_objs = self._done_objs
        self._done_objs = []
        self._objects = []
        self.point = None
        self.normal = None
        self.target = None

    def register(self):
        if self.collision not in self.game_object.collisionCallbacks:
            self.game_object.collisionCallbacks.append(self.collision)
        if self.reset not in logic.getCurrentScene().pre_draw:
            logic.getCurrentScene().pre_draw.append(self.reset)

    def remove(self):
        self.game_object.collisionCallbacks.remove(self.collision)
        logic.getCurrentScene().pre_draw.remove(self.reset)


class ULCollision(Collision):
    _deprecated = True


def on_collision(
    obj: GameObject,
    callback: Callable,
    prop: str = '',
    material: str = '',
    tap: bool = False,
    post_call: bool = False
) -> Collision:
    """Bind a callback to an object's collision detection.

    :param obj: Object whose collision detection will be monitored.
    :param callback: Callback to be called when collision occurs. Must have arguments `(obj, point, normal)`.
    :param prop: Only look for objects that have this property.
    :param material: Only look for objects that have this material applied.
    :param tap: Only validate the first frame of the collision.
    """
    if not isinstance(obj, GameObject):
        error("'on_collision()' Argument 0: Expected 'KX_GameObject' type!")
        return
    return Collision(obj, callback, prop, material, tap, post_call)



from uplogic.nodes import ULActionNode
from uplogic import console
import bpy
import json
import os


class ULRemoveVariable(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.name = None
        self.file_name = None
        self.path = ''
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def write_to_json(self, path, name):
        data = None
        if not path.endswith('.json'):
            path = os.path.join(path, f'{self.get_input(self.file_name)}.json')
        if os.path.isfile(path):
            f = open(path, 'r')
            data = json.load(f)
            if name in data:
                del data[name]
            f.close()
            f = open(path, 'w')
            json.dump(data, f, indent=2)
            f.close()
        else:
            console.debug('File does not exist!')

    def evaluate(self):
        if not self.get_condition():
            return
        name = self.get_input(self.name)
        path = self.get_input(self.path)

        path = (bpy.path.abspath(path))
        os.makedirs(path, exist_ok=True)

        self.write_to_json(path, name)
        self._done = True



from .shader import Filter2D


glsl = """
uniform sampler2D bgl_RenderedTexture;
// in vec4 bgl_TexCoord;
uniform float brightness;

// out vec4 fragColor;

void main() {
    fragColor = texture(bgl_RenderedTexture, bgl_TexCoord.xy) * brightness;
}
"""


class Brightness(Filter2D):

    def __init__(self, brightness=1.0, idx: int = None) -> None:
        self.uniforms = {'brightness': float(brightness)}
        super().__init__(glsl, idx, {'brightness': self.uniforms})

    @property
    def brightness(self):
        return self.uniforms['brightness']

    @brightness.setter
    def brightness(self, val):
        self.uniforms['brightness'] = float(val)



from bge.types import KX_GameObject
from uplogic.nodes import ULConditionNode
from uplogic.physics.collision import Collision
from uplogic.nodes import ULLogicContainer
from bpy.types import Material
from mathutils import Vector


class ULCollision(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.game_object = None
        self.use_mat = False
        self.prop = ''
        self.material = ''
        self.pulse = False
        self._collision = None
        self._active = None
        self._point = None
        self._normal = None
        self._target = None
        self._objects = None

        # self._target = None
        # self._point = None
        # self._normal = None
        # self._collision_triggered = False
        # self._consumed = False
        # self._game_object = None
        # self._objects = []

        self.COLLISION = self.add_output(self.get_collision)
        self.TARGET = self.add_output(self.get_target)
        self.POINT = self.add_output(self.get_point)
        self.NORMAL = self.add_output(self.get_normal)
        self.OBJECTS = self.add_output(self.get_objects)
        
    def get_collision(self):
        return self._collision.active

    def get_point(self):
        return self._collision.point

    def get_normal(self):
        return self._collision.normal

    def get_target(self):
        return self._collision.target

    def get_objects(self):
        return self._collision._objects

    def on_collision(self, obj, point, normal):
        pass

    def evaluate(self):
        game_obj = self.get_input(self.game_object)
        if self._collision is None or self._collision.game_object is not game_obj:
            self._collision = Collision(
                game_obj,
                self.on_collision,
                self.get_input(self.prop),
                self.get_input(self.material),
                not self.get_input(self.pulse),
                True
            )



from uplogic.nodes import ULActionNode


class ULApplyImpulse(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.point = None
        self.impulse = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        point = self.get_input(self.point)
        impulse = self.get_input(self.impulse)
        local = self.local
        if impulse:
            game_object.applyImpulse(point, impulse, local)
        self._done = True



from uplogic.nodes import ULParameterNode
from uplogic.utils import lerp


class ULInterpolate(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.a = None
        self.b = None
        self.fac = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        a = self.get_input(self.a)
        b = self.get_input(self.b)
        fac = self.get_input(self.fac)
        return lerp(a, b, fac)



from uplogic.nodes import ULParameterNode
from uplogic.utils.constants import LO_AXIS_TO_VECTOR


class ULAxisVector(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.OUT = self.add_output(self.get_vec)

    def get_vec(self):
        obj = self.get_input(self.game_object)
        front_vector = LO_AXIS_TO_VECTOR[self.axis]
        return obj.getAxisVect(front_vector)



__version__ = '5.0.0'

try:
    from . import console
    from . import animation
    from . import audio
    from . import data
    from . import events
    from . import input
    from . import network
    from . import physics
    from . import shaders
    from . import ui
    from . import utils
except:
    print('Not in runtime!')
import bpy


def check_version(version: str):
    version = str(version)
    nums = version.split('.')
    vnums = __version__.split('.')
    
    while len(vnums) < len(nums):
        vnums.append(0)
    
    for i, num in enumerate(nums):
        if int(num) != int(vnums[i]):
            return False

    return True





from uplogic.nodes import ULOutSocket, ULParameterNode


class ULGetOwner(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.owner = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        self.owner



from .shader import Filter2D
import bpy, bge
from mathutils import Vector
from ..console import error


class Texture(Filter2D):

    glsl = """
uniform sampler2D bgl_RenderedTexture;
uniform sampler2D tex;
uniform float opacity;
uniform vec2 pos;
uniform vec2 size;

in vec4 bgl_TexCoord;

out vec4 fragColor;


float map_range(float value, float in_min, float in_max, float out_min, float out_max){
    return (value - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}


void main()
{

    vec2 texcoord = bgl_TexCoord.xy;
    float opa = opacity;

    if (texcoord.x < pos.x || texcoord.x > pos.x + size.x || texcoord.y < pos.y || texcoord.y > pos.y + size.y){
        opa = 0;
    }

    vec2 texMap = vec2(
        map_range(texcoord.x, pos.x, pos.x + size.x, 0, 1),
        map_range(texcoord.y, pos.y, pos.y + size.y, 0, 1)
    );
    fragColor = mix(
        texture(bgl_RenderedTexture, texcoord),
        texture(tex, texMap),
        opa
    );
}
"""

    def __init__(self, texture: bpy.types.Image = None, opacity: float = 1.0, pos=Vector((0, 0)), size=Vector((1, 1)), idx: int = None) -> None:
        texture = bpy.data.images.get(str(texture), texture)
        if not isinstance(texture, bpy.types.Image):
            error("'Texture': first argument requires an object of type 'bpy.types.Image'!")
            return
        texture.gl_load()
        self.free_textures = True
        self.uniforms = {'tex': texture, 'opacity': float(opacity), 'pos': Vector(pos), 'size': Vector(size)}
        super().__init__(self.glsl, idx, {'tex': self.uniforms, 'opacity': self.uniforms, 'pos': self.uniforms, 'size': self.uniforms})

    @property
    def texture(self):
        return self.uniforms['tex']

    @texture.setter
    def texture(self, val):
        if self.free_textures:
            self.free_texture()
        if not isinstance(val, bpy.types.Image):
            raise TypeError
        if not val.has_data:
            val.gl_load()
        self.uniforms['tex'] = val

    def free_texture(self):
        self.uniforms['tex'].gl_free()
        self.uniforms['tex'].buffers_free()

    @property
    def opacity(self):
        return self.uniforms['opacity']

    @opacity.setter
    def opacity(self, val):
        self.uniforms['opacity'] = float(val)

    @property
    def pos(self):
        return self.uniforms['pos']

    @pos.setter
    def pos(self, val):
        self.uniforms['pos'] = Vector(val).to_2d()

    @property
    def size(self):
        return self.uniforms['size']

    @size.setter
    def size(self, val):
        self.uniforms['size'] = Vector(val).to_2d()


class Mask(Texture):

    glsl = """
uniform sampler2D bgl_RenderedTexture;
uniform float bgl_RenderedTextureWidth;
uniform float bgl_RenderedTextureHeight;
vec2 resolution = vec2(bgl_RenderedTextureWidth, bgl_RenderedTextureHeight);

uniform sampler2D tex;
uniform float opacity;
uniform vec2 pos;
uniform vec2 size;
uniform float threshold;
uniform int blur_samples;
uniform int blur_quality;
uniform int blur_radius;

in vec4 bgl_TexCoord;

out vec4 fragColor;


float map_range(float value, float in_min, float in_max, float out_min, float out_max){
    return (value - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}


void main()
{
    float Pi = 6.28318530718;

    vec2 texcoord = bgl_TexCoord.xy;
    float opa = opacity;

    if (texcoord.x < pos.x || texcoord.x > pos.x + size.x || texcoord.y < pos.y || texcoord.y > pos.y + size.y){
        opa = 0;
    }

    vec2 texMap = vec2(
        map_range(texcoord.x, pos.x, pos.x + size.x, 0, 1),
        map_range(texcoord.y, pos.y, pos.y + size.y, 0, 1)
    );

    vec4 mask = mix(vec4(1.0), texture(tex, texMap), opa);

    vec2 radius = blur_radius / resolution.xy;

    for (float d = 0.0; d < Pi; d += Pi / 15)
    {
		for(float i = 1.0; i <= 1.0; i += 1.0)
        {
			mask += texture(tex, texcoord - vec2(cos(d), sin(d)) * radius * i);
        }
    }
    mask /= blur_samples;

    mask *= vec4(1 + threshold);
    mask -= vec4(threshold);

    fragColor = texture(bgl_RenderedTexture, texcoord) * clamp(mask, 0.0, 1.0);
}
"""

    def __init__(
        self,
        texture: bpy.types.Image = None,
        opacity: float = 1.0,
        pos=Vector((0, 0)),
        size=Vector((1, 1)),
        threshold=0.0,
        blur_samples=15,
        blur_radius=0.003,
        idx: int = None
    ) -> None:
        texture = bpy.data.images.get(str(texture), texture)
        if not isinstance(texture, bpy.types.Image):
            error("'Texture': first argument requires an object of type 'bpy.types.Image'!")
            return
        texture.gl_load()
        self.free_textures = True
        super().__init__(texture=texture, opacity=opacity, pos=pos, size=size, idx=idx)
        self.uniforms.update({
            'blur_samples': int(blur_samples),
            'blur_radius': int(blur_radius),
            'threshold': float(threshold)
        })
        self._uniforms.update({
            'blur_samples': self.uniforms,
            'blur_radius': self.uniforms,
            'threshold': self.uniforms,
            'blur_quality': self.uniforms
        })

    @property
    def blur_radius(self):
        return self.uniforms['blur_radius']

    @blur_radius.setter
    def blur_radius(self, val):
        self.uniforms['blur_radius'] = int(val)

    @property
    def blur_samples(self):
        return self.uniforms['blur_samples']

    @blur_samples.setter
    def blur_samples(self, val):
        self.uniforms['blur_samples'] = int(val)

    @property
    def threshold(self):
        return self.uniforms['threshold']

    @threshold.setter
    def threshold(self, val):
        self.uniforms['threshold'] = float(val)



from uplogic.animation import ActionSystem
from uplogic.data import GlobalDB
from uplogic.nodes import ULParameterNode


class ULActionStatus(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.action_layer = None
        self._action_name = None
        self._action_frame = None
        self.action = None
        self._playing = False
        self.act_system = self.get_act_system()
        self.OUT = self.add_output(self.get_out)
        self.ACTION_NAME = self.add_output(self.get_action_name)
        self.ACTION_FRAME = self.add_output(self.get_action_frame)

    def fetch(self):
        game_object = self.get_input(self.game_object)
        action_layer = self.get_input(self.action_layer)
        self.action = self.act_system.get_layer(game_object, action_layer)

    def get_out(self):
        return bool(self.action and self.action.is_playing)

    def get_act_system(self):
        act_systems = GlobalDB.retrieve('uplogic.animation')
        if act_systems.check('default'):
            return act_systems.get('default')
        else:
            return ActionSystem('default')

    def get_action_name(self):
        return self.action.name if self.action else None

    def get_action_frame(self):
        return self.action.frame if self.action else -1



from uplogic.nodes import ULConditionNode
from uplogic.utils.constants import LOGIC_OPERATORS
from mathutils import Vector


class ULCheckDistance(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.operation = 0
        self.param_a = None
        self.param_b = None
        self.dist = None
        self._distance = 0
        self.OUT = self.add_output(self.get_result)
        self.DIST = self.add_output(self.get_distance)

    def get_result(self):
        dist = self.get_input(self.dist)
        return LOGIC_OPERATORS[self.operation](self._distance, dist)

    def get_distance(self):
        return self._distance

    def evaluate(self):
        self._distance = (Vector(self.get_input(self.param_a)) - Vector(self.get_input(self.param_b))).length
